
## Copyright (c) 2009-2010, Scott D. Peckham
## January 2009  (converted from IDL)
## December 2009 (starting from ET_priestley_taylor.py)
## May 2010 (changes to unit_test() and read_cfg_file()

#################################################
#  NB!  See note at "THIS MAY BE COSTLY"
#################################################

#-----------------------------------------------------------------------
#  NOTES:  This file defines a "Read from file" ET component
#          and related functions.  It inherits from the ET
#          "base class" in "ET_base.py".
#-----------------------------------------------------------------------
#
#  unit_test()
#
#  class ET_component
#
#      get_attribute()         # (10/26/11)
#      get_input_items()
#      get_output_items()
#---------------------------
#      check_input_types()
#      update_ET_rate()
#      update_water_balance()   ###### Overrides one in ET_base.py
#      open_input_files()
#      read_input_files()
#      close_input_files()
#
#  Functions:
#      RTG_to_RTS()
#
#-----------------------------------------------------------------------

from numpy import *
import numpy

import os

import cfg_files as cfg
import ET_base
import model_input
import tf_utils

#-----------------------------------------------------------------------
def unit_test():

    ec = ET_component()
    ec.CCA   = False
    ec.DEBUG = False
    
    #-----------------------------------------
    # This function adjusts for the platform
    # and can be changed in "tf_utils.py".
    #-----------------------------------------
    cfg_directory   = tf_utils.TF_Test_Directory()
    os.chdir( cfg_directory )
    cfg_prefix      = tf_utils.TF_Test_Case_Prefix()
    ec.site_prefix  = 'Treynor'  #############  
  
    ec.run_model( cfg_directory=cfg_directory,
                  cfg_prefix=cfg_prefix )

##    ec.initialize( cfg_prefix=cfg_prefix, mode="driver" )
 
##    print 'nx =', ec.nx
##    print 'ny =', ec.ny
    
#   unit_test()
#-----------------------------------------------------------------------
class ET_component(ET_base.ET_component):

    #-------------------------------------------------------------------
    def get_attribute(self, att_name):

        map = {'comp_name':          'EvapReadFile',
               'version':            '3.1',
               'model_name':         'Evaporation_Read_File',
               'model_family':       'TopoFlow',
               'cfg_template_file':  'Evap_Read_File.cfg.in',
               'cfg_extension':      '_evap_read_file.cfg',
               'cmt_var_prefix':     '/EvapReadFile/Input/Var/',
               'gui_xml_file':       '/home/csdms/cca/topoflow/3.1/src/share/cmt/gui/Evap_Read_File.xml',
               'dialog_title':       'Evaporation: Read File Parameters',
               'time_step_type':     'fixed',
               'time_units':         'seconds',
               'mesh_type':          'none',
               'author_name':        'Scott Peckham'}

        try:
            return map[ att_name.lower() ]
        except:
            print '###################################################'
            print ' ERROR: Could not find attribute: ' + att_name
            print '###################################################'
            print ' '

    #   get_attribute() 
    #-------------------------------------------------------------------
    def get_input_items(self):

        items = ['d']
        return items
        ## return numpy.array(items)    # (string array vs. list)
    
    #   get_input_items()
    #-------------------------------------------------------------------
    def get_output_items(self):

        items = ['dt', 'da', 'ET', 'vol_ET']
        return items
        ## return numpy.array(items)    # (string array vs. list)
    
    #   get_output_items()
    #-------------------------------------------------------------------
##    def check_input_types(self):
##
##        #----------------------------------------------------
##        # Notes: Usually this will be overridden by a given
##        #        method of computing ET.
##        #----------------------------------------------------
##        are_scalars = numpy.array([
##                         self.is_scalar('T_soil_x'),
##                         self.is_scalar('soil_x'),
##                         self.is_scalar('K_soil'),
##                         self.is_scalar('alpha'),
##                         #-------------------------------
##                         self.mp.is_scalar('Qn_SW'),
##                         self.mp.is_scalar('Qn_LW'),
##                         self.mp.is_scalar('T_air'),
##                         self.mp.is_scalar('T_surf'),
##                         #-------------------------------
##                         self.cp.is_scalar('d'),
##                         #-------------------------------
##                         self.gp.is_scalar('h_table') ])
##
##        self.ALL_SCALARS = numpy.all(are_scalars)
##        
##    #   check_input_types()
    #-------------------------------------------------------------------
    def update_ET_rate(self):

        #----------------------------------------------------------
        # Note:  This function should return without doing
        #        anything.  For this component, the ET rate
        #        is not computed but is read directly as the
        #        only "input" variable.  Note that "initialize()"
        #        calls "read_input_files()" so that ET has been
        #        defined before first call to ET_base.update().
        #        Recall that ET_base.update() will call
        #        read_input_files() to get next ET grid.
        #----------------------------------------------------------
        #        We may want to make sure ET values read from
        #        input file are positive before using them.
        #        If so, uncomment 2 lines below.
        #----------------------------------------------------------
        return 
##        if (self.ET != None):
##            self.ET = maximum(self.ET, float64(0))

        ##########################################
        #  THIS MAY BE COSTLY.  BETTER WAY OR
        #  ALLOW ET TO BE A SCALAR ??
        ##########################################
##        if (size(self.ET) == 1):
##            self.ET += zeros((self.ny, self.nx), dtype='Float64')
            
    #   update_ET_rate()
    #-------------------------------------------------------------------  
    def update_water_balance(self):

        #----------------------------------------------------------
        # Notes: Return without modifying the surface water depth
        #        or soil moisture in the top soil layer.
        #        Depending on how the model is run, this may
        #        lead to an inconsistent result where the mass
        #        of water in the system is not conserved.
        #----------------------------------------------------------
        print '-----------------------------------------------------'
        print 'WARNING: ET_read_file component assumes that the'
        print '         ET rates read from files are actual vs.'
        print '         potential rates.  It does not consume'
        print '         surface or subsurface water.'
        print '-----------------------------------------------------'
        return
    
    #   update_water_balance()
    #-------------------------------------------------------------------  
    def open_input_files(self):

        self.ET_file = self.in_directory + self.ET_file

        self.ET_unit = model_input.open_file(self.ET_type, \
                                             self.ET_file)
        print 'self.ET_type =', self.ET_type
        print 'self.ET_file =', self.ET_file
        
##        self.duration_unit  = model_input.open_file(self.duration_type,  \
##                                                    self.duration_file)
##        print 'self.duration_type (ET) =', self.duration_type
##        print 'self.duration_file (ET) =', self.duration_file

    #   open_input_files()
    #-------------------------------------------------------------------  
    def read_input_files(self):

        #--------------
        # For testing
        #--------------
        ## print 'self.ET_type =', self.ET_type
        ## print 'self.ET_file =', self.ET_file
        
        rti = self.rti
        
        #-------------------------------------------------------
        # All grids are assumed to have a data type of Float32.
        #-------------------------------------------------------
        ET = model_input.read_next(self.ET_unit, self.ET_type, rti,
                                   factor=self.mmph_to_mps)
        if (ET != None):
            self.ET = ET
            print 'min(ET) =', ET.min() * self.mps_to_mmph, ' [mmph]'
            print 'min(ET) =', ET.max() * self.mps_to_mmph, ' [mmph]'
##            print 'min(ET) =', ET.min(), ' [mps]'
##            print 'max(ET) =', ET.max(), ' [mps]'
            
    #   read_input_files()
    #-------------------------------------------------------------------
    def close_input_files(self):
 
        try:
            self.ET_unit.close()
        except:
            print 'Could not close input file "ET_unit".'
            
        ## if (self.ET_file != ''): self.ET_unit.close()
        
    #   close_input_files()
    #-------------------------------------------------------------------
#-----------------------------------------------------------------------
def RTG_to_RTS( RTG_prefix, new_RTS_file,
                nx=None, ny=None, dtype='float32' ):

    #--------------------------------------------------------------
    # Notes:  This function "bundles" a set of binary grid files
    #         (generic RTG format) with the same file_name prefix
    #         and embedded time index into a single RTS file.
    #--------------------------------------------------------------
    import glob
    import os.path
    import numpy
    import rti_files

    #--------------------
    # Open new RTS file
    #--------------------
    RTS_EXISTS = os.path.exists( new_RTS_file )
    if (RTS_EXISTS):
        print 'SORRY, An RTS file with the name'
        print   new_RTS_file
        print 'already exists.'
        return

    RTG_list = glob.glob( RTG_prefix + '*' )
    RTG_list = numpy.sort( RTG_list )
    print 'Number of RTG files to merge =', len(RTG_list)
    
    #--------------------------------
    # Try to get info from RTI file
    #--------------------------------
    if (nx == None) and (ny == None):
        RTG_file1 = RTG_list[0]
        RTI_file  = rti_files.try_to_find_rti_file( RTG_file1 )
        if (RTI_file != 'none'):
            info = rti_files.read_info( RTI_file )
            nx   = info.ncols
            ny   = info.nrows
        else:
            print 'ERROR: Could not find RTI file and nx and ny not provided.'
            return
##    else:
##        #--------------------------------------------
##        # For case when used as a script under Unix
##        # (used in the file rtg2rts.py)
##        #--------------------------------------------
##        nx = eval(nx)
##        ny = eval(ny)
        
    RTS_unit = open( new_RTS_file, 'wb' )

    #------------------------------
    # Write RTG grids to RTS file
    #------------------------------    
    for RTG_file in RTG_list:
        print 'Reading values from:', RTG_file
        RTG_unit = open( RTG_file, 'rb')
        grid     = numpy.fromfile( RTG_unit, count=nx*ny, dtype=dtype )
        RTG_unit.close()
        grid.tofile( RTS_unit )

    #---------------------
    # Close new RTS file
    #---------------------
    RTS_unit.close()
    print 'Finished writing RTG files to RTS format.'
    print ' '
    
#   RTG_to_RTS()
#-----------------------------------------------------------------------
    


