
## Copyright (c) 2001-2009, Scott D. Peckham
## January 2009  (converted from IDL)
## May, July, August 2009

#-----------------------------------------------------------------------
#  NB!     In the CFG file, change MANNING and LAW_OF_WALL flags to
#          a single string entry like "friction method".   #########
#-----------------------------------------------------------------------
#  Notes:  Set self.u in manning and law_of_wall functions ??
#          Update friction factor in manning() and law_of_wall() ?
#          Double check how Rh is used in law_of_the_wall().

#          d8_flow has "flow_grids", but this one has "codes".
#          Make sure values are not stored twice.
#-----------------------------------------------------------------------

#-----------------------------------------------------------------------
#  NOTES:  This file defines a "base class" for channelized flow
#          components as well as functions used by most or
#          all channel flow methods.  The methods of this class
#          (especially "update_velocity") should be over-ridden as
#          necessary for different methods of modeling channelized
#          flow.  See channels_kinematic_wave.py,
#          channels_diffusive_wave.py and channels_dynamic_wave.py.
#-----------------------------------------------------------------------
#
#  unit_test()
#
#  class channels_component
#      __init__()
#-----------------------------
#      initialize()
#      update()
#      finalize()
#      read_config_file()    #### OBSOLETE HERE SOON #####
#-----------------------------
#      get_cca_ports()           # (in CSDMS_base.py)
#      embed_child_components()
#      add_child_ports()
#      initialize_ports()
#      release_cca_ports()       # (in CSDMS_base.py)
#----------------------------------
#      initialize_d8_vars()          ########
#      initialize_computed_vars()
#      initialize_outlet_values()
#      initialize_peak_values()
#----------------------------------
#      update_R()
#      update_R_integral()
#      update_discharge()
#      update_diversions()          # (Feb. 16, 2010)
#      update_flow_volume()
#      update_flow_depth()
#      update_free_surface_slope()
#      update_trapezoid_Rh()
#      update_velocity()            # (override as needed)
#      update_velocity_on_edges()
#      update_outlet_values()
#      update_peak_values()         # (at the main outlet)
#      update_volume_out()          # (moved here from basins.py)
#      update_mins_and_maxes()      # (don't add into update())
#      check_flow_depth()
#      check_flow_velocity()
#----------------------------------
#      open_input_files()
#      read_input_files()
#      close_input_files()
#----------------------------------
#      update_outfile_names()
#      open_output_files()
#      write_output_files()
#      close_output_files()
#      save_grids()
#      save_pixel_values()
#----------------------------------
#      manning_formula()
#      law_of_the_wall()
#      print_status_report()
#      remove_bad_slopes() 

#  Functions:               # (stand-alone versions of these)
#      Trapezoid_Rh()
#      Manning_Formula()
#      Law_of_the_Wall()
    
#-----------------------------------------------------------------------

from numpy import *
import numpy

import cfg_files as cfg
import CSDMS_base
# import d8_base
import model_input
import rtg_files
import tf_d8_base as d8_base
import tf_utils

# from save_load import *  # (for Read_Vars, etc)
from model_output import *

#-----------------------------------------------------------------------
def unit_test():
 
    directory   = tf_utils.TF_Test_Directory()
    data_prefix = tf_utils.TF_Test_Data_Prefix()
    case_prefix = tf_utils.TF_Test_Case_Prefix()
  
    c = channels_component()
    c.CCA   = False
    c.DEBUG = False
    
    #-------------------------------------------
    # Call initialize() and call update() once
    #-------------------------------------------
##    print 'STATUS =', c.get_status()
##    c.initialize(directory=directory,
##                 data_prefix=data_prefix,
##                 case_prefix=case_prefix, mode="main")
##    print 'STATUS =', c.get_status()
##    time_sec = float64(0)
##    c.update(time_sec)
##    print 'STATUS =', c.get_status()

    c.run_model(directory=directory,
                data_prefix=data_prefix,
                case_prefix=case_prefix)

    print ' '
    print '------------------------------------------------------'
    print ' WARNING!  Running unit_test() for channels_base.py'
    print '           so all velocities will be zero.'
    print '------------------------------------------------------'
    print ' '
    
##    print '---------------------------------------'
##    P = c.get_port_data('P', c.mp)
##    print ' '
##    print 'Finished with CHANNELS.run_model().'
##    print '---------------------------------------'
##    print 'Reading from "met" port:'
##    print 'P =', P
##    print ' '
    
##    print 'KINEMATIC WAVE =', c.KINEMATIC_WAVE
##    print 'DIFFUSIVE WAVE =', c.DIFFUSIVE_WAVE
##    print 'DYNAMIC WAVE   =', c.DYNAMIC_WAVE
##    print ' '
##    print 'nx          =', c.nx
##    print 'ny          =', c.ny
##    print 'config_file =', c.config_file
    
#   unit_test()
#-----------------------------------------------------------------------
class channels_component(CSDMS_base.CSDMS_component):

    #-----------------------------------------------------------
    # Notes: This is first called without the prefix argument.
    #        The grid filenames below are set using the data
    #        prefix by the Read_Run_Info_Panel routine.

    # NB!   (3/16/07) dt is now set to zero initially, so
    #        that Read_Run_Info_Panel can determine if user
    #        has set the value (e.g. by loading a saved value).
    #        If unset, then it is set by Read_Run_Info_Panel.
    #        If GUI is not used, it is set in the input file.
    #-----------------------------------------------------------
    
    #-------------------------------------------------------------------
    def set_constants(self):

        #------------------------
        # Define some constants
        #------------------------
        self.g          = float64(9.81)    # (gravitation const.)
        self.aval       = float64(0.476)   # (integration const.)
        self.kappa      = float64(0.408)   # (von Karman's const.)
        self.law_const  = sqrt(self.g) / self.kappa
        self.one_third  = float64(1.0) / 3.0        
        self.two_thirds = float64(2.0) / 3.0
        self.deg_to_rad = numpy.pi / 180.0
        
    #   set_constants()
    #-------------------------------------------------------------------
    def get_input_items(self):

        items = ['P', 'SM', 'ET', 'IN', 'GW']
        return items
        ## return numpy.array(items)    # (string array vs. list)

    #   get_input_items()
    #-------------------------------------------------------------------
    def get_output_items(self):

        items = ['dt','da','d0','R','vol_R','volume_out',
                 'Q_peak', 'u_peak', 'd_peak',
                 'T_peak', 'Tu_peak', 'Td_peak',
                 'Q_outlet', 'u_outlet', 'd_outlet', 'f_outlet',
                 # 'Q_min', 'Q_max', 'u_min', 'u_max', 'd_min', 'd_max',
                 'time_sec', 'time_min',
                 #-------------------------------------------------
                 'Q', 'u', 'd', 'f', 'vol', 'R', 'Rh',
                 'widths', 'angles', 'nvals', 'z0vals', 'sinu',
                 'slopes', 'S_bed', 'S_free',
                 'nvals_min', 'nvals_max', 'z0vals_min', 'z0vals_max']

        return items
        ## return numpy.array(items)    # (string array vs. list)
    
    #   get_output_items()
    #-------------------------------------------------------------------
    def initialize(self, directory=None,
                   data_prefix=None,
                   case_prefix=None, mode="module"):

        self.status = 'initializing'  # (OpenMI 2.0 convention)
        self.mode   = mode
        print 'Channels component: Initializing...'
        self.set_directory(directory, data_prefix, case_prefix)

        #-----------------------------------------------
        # Load component parameters from a config file
        #-----------------------------------------------
        self.set_constants()      # (12/7/09)
        self.read_config_file() 
        
        if (self.method == 0):
            self.SAVE_Q_PIXELS = False    # (It is True by default.)
            self.DONE = True
            self.status = 'initialized'  # (OpenMI 2.0 convention) 
            return

        self.read_grid_info()
             
        #---------------------------------------------
        # Open input files needed to initialize vars 
        #---------------------------------------------
        # Can't move read_input_files() to start of
        # update(), since initial values needed here.
        #---------------------------------------------
        self.open_input_files()
        self.read_input_files()

        #-----------------------
        # Initialize variables
        #-----------------------
        self.initialize_d8_vars()  # (depend on D8 flow grid)
        self.initialize_computed_vars()
        self.initialize_time_vars()
  
        #--------------------------------------------------
        # Make sure self.Q_ts_file is not NULL (12/22/05)
        
        # This is only output file that is set by default
        # and is still NULL if user hasn't opened the
        # output var dialog for the channel process.
        #--------------------------------------------------
        if (self.SAVE_Q_PIXELS and (self.Q_ts_file == '')):    
            self.Q_ts_file = (case_prefix + '_0D-Q.txt')       

        ###############################################################
        ###############################################################
        #  If this component is running in stand-alone mode,
        #  then it must initialize the process modules that
        #  it is going to use.  If it is being called by another
        #  component, should caller pass ports in via "set_value" ??
        ###############################################################
        ###############################################################

        #----------------------------------------------------------
        # Channels component is initialized before Diversions
        # component, which may override these settings. (2/16/10)
        #----------------------------------------------------------
        self.use_sources = False
        self.use_sinks   = False
        self.use_canals  = False
        
        self.initialize_required_components(mode)
            
        #-------------------------------
        # Save these for writing output
        #-------------------------------
        self.store_outlet_IDs()
        
        self.open_output_files()
        
        self.status = 'initialized'  # (OpenMI 2.0 convention) 
        
    #   initialize()
    #-------------------------------------------------------------------
    def update(self, time_seconds=None):

        #---------------------------------------------
        # Note that u and d from previous time step
        # must be used on RHS of the equations here.
        #---------------------------------------------
        self.status = 'updating'  # (OpenMI 2.0 convention)
        if (self.mode == 'main'):
            self.print_time_and_value(self.Q_outlet, 'Q_out', '[m^3/s]')
        
        #-------------------------
        # Update computed values
        #-------------------------
        self.update_R()       
        self.update_R_integral()
        self.update_discharge()
        self.update_diversions()     ### (Feb. 16, 2010)
        self.update_flow_volume()
        self.update_flow_depth()
        if not(self.DYNAMIC_WAVE):
            self.update_trapezoid_Rh()
            # print 'Rhmin, Rhmax =', self.Rh.min(), self.Rh.max()
        self.update_velocity()
        self.update_velocity_on_edges()     # (set to zero)
##        print 'Rmin, Rmax =', self.R.min(), self.R.max()
##        print 'Qmin,  Qmax =',  self.Q.min(), self.Q.max()
##        print 'umin,  umax =',  self.u.min(), self.u.max()
##        print 'dmin,  dmax =',  self.d.min(), self.d.max()
##        print 'nmin,  nmax =',  self.nvals.min(), self.nvals.max()
##        print 'Rhmin, Rhmax =', self.Rh.min(), self.Rh.max()
##        print 'Smin,  Smax =',  self.S_bed.min(), self.S_bed.max()
        self.update_outlet_values()
        self.update_peak_values()
        self.update_volume_out()    # (rename to update_Q_out ??)
        
        #------------------------
        # Check computed values
        #------------------------
        D_OK = self.check_flow_depth()
        U_OK = self.check_flow_velocity()
        OK   = (D_OK and U_OK)

        #-------------------------------------------
        # Read from files as needed to update vars 
        #-----------------------------------------------------
        # NB! This is currently not needed for the "channel
        # process" because values don't change over time and
        # read_input_files() is called by initialize().
        #-----------------------------------------------------
        # if (self.time_index > 0):
        #     self.read_input_files()

        #----------------------------------------------
        # Write user-specified data to output files ?
        #----------------------------------------------
        self.write_output_files(time_seconds)
        if (OK):
            self.status = 'updated'  # (OpenMI 2.0 convention)
        else:
            self.status = 'failed'
            self.DONE   = True

        #------------------------
        # Update internal clock
        #------------------------
        self.update_time()
        
    #   update()   
    #-------------------------------------------------------------------
    def finalize(self):

        self.status = 'finalizing'  # (OpenMI)   
        self.close_input_files()   ##  TopoFlow input "data streams"
        self.close_output_files()
        self.status = 'finalized'  # (OpenMI)
        print 'Channels component: Finished.'

        #---------------------------
        # Release all of the ports
        #----------------------------------------
        # Make this call in "finalize()" method
        # of the component's CCA Imple file
        #----------------------------------------
        # self.release_cca_ports( port_names, d_services )
        
    #   finalize()
    #-------------------------------------------------------------------
    def read_config_file(self):

        #------------------------------------------
        # Read parameters from a CFG file that is
        # in the current working directory.
        #------------------------------------------
        print 'Channels component: Reading config file...'
        self.config_file = (self.directory +
                            self.case_prefix + '_channels_kinematic_wave.cfg')
        file_unit = open(self.config_file, 'r')
        
        #-------------------------
        # Skip over header lines
        #-------------------------
        cfg.skip_header( file_unit, n_lines=4 )
            
        #------------------------
        # Read the channel vars
        #------------------------
        self.method      = cfg.read_value( file_unit, dtype='int16' )
        self.method_name = cfg.read_value( file_unit, dtype='string' )
        #-----------------------------------------------------------------
        self.MANNING     = cfg.read_value( file_unit, dtype='boolean' )
        self.LAW_OF_WALL = cfg.read_value( file_unit, dtype='boolean' )
        #-----------------------------------------------------------------        
        dt_info          = cfg.read_input_option( file_unit )
        # self.dt_type   = dt_info[0]
        self.dt          = dt_info[1]
        #-----------------------------------------------------------------
        self.KINEMATIC_WAVE = ("kinematic" in self.method_name.lower())
        self.DIFFUSIVE_WAVE = ("diffusive" in self.method_name.lower())
        self.DYNAMIC_WAVE   = ("dynamic"   in self.method_name.lower())
        #-----------------------------------------------------------------
        code_info  = cfg.read_input_option( file_unit )
        slope_info = cfg.read_input_option( file_unit )
        if (self.MANNING):    
            nval_info  = cfg.read_input_option( file_unit )
        if (self.LAW_OF_WALL):    
            z0val_info = cfg.read_input_option( file_unit )
        width_info = cfg.read_input_option( file_unit )
        angle_info = cfg.read_input_option( file_unit )
        d0_info    = cfg.read_input_option( file_unit )
        sinu_info  = cfg.read_input_option( file_unit )
        #-----------------------------------------------------------
        self.code_type   = code_info[0]
        self.codes       = code_info[1]
        self.code_file   = code_info[2]
        #-----------------------------------------------------------
        self.slope_type  = slope_info[0]
        self.slopes      = slope_info[1]
        self.slope_file  = slope_info[2]
        #-----------------------------------------------------------
        if (self.MANNING):
            self.nval_type   = nval_info[0]
            self.nvals       = nval_info[1]
            self.nval_file   = nval_info[2]
            self.z0val_file  = ''  #### need this
            if (self.nvals != None):
                self.nvals_min = self.nvals.min()  ######
                self.nvals_max = self.nvals.max()  ######
        #-----------------------------------------------------------
        if (self.LAW_OF_WALL):
            self.z0val_type = z0val_info[0]
            self.z0vals     = z0val_info[1]
            self.z0val_file = z0val_info[2]
            self.nval_file   = ''  #### need this
            if (self.z0vals != None):
                self.z0vals_min = self.z0vals.min()  ######
                self.z0vals_max = self.z0vals.max()  ######
        #-----------------------------------------------------------
        self.width_type  = width_info[0]
        self.widths      = width_info[1]
        self.width_file  = width_info[2]
        #-----------------------------------------------------------
        self.angle_type  = angle_info[0]
        self.angles      = angle_info[1]
        self.angle_file  = angle_info[2]
        #-----------------------------------------------------------
        self.d0_type     = d0_info[0]
        self.d0          = d0_info[1]
        self.d0_file     = d0_info[2]
        #-----------------------------------------------------------
        self.sinu_type   = sinu_info[0]
        self.sinu        = sinu_info[1]
        self.sinu_file   = sinu_info[2]
        #---------------------------------------------------------------
        #---------------------------------------------------------------
        save_grid_dt_info       = cfg.read_input_option( file_unit )
        save_Q_grids, Q_gs_file = cfg.read_output_option( file_unit )
        save_u_grids, u_gs_file = cfg.read_output_option( file_unit )
        save_d_grids, d_gs_file = cfg.read_output_option( file_unit )
        save_f_grids, f_gs_file = cfg.read_output_option( file_unit )
        #---------------------------------------------------------------
        self.save_grid_dt = save_grid_dt_info[1]
        self.SAVE_Q_GRIDS = save_Q_grids
        self.SAVE_U_GRIDS = save_u_grids
        self.SAVE_D_GRIDS = save_d_grids
        self.SAVE_F_GRIDS = save_f_grids
        self.Q_gs_file    = Q_gs_file
        self.u_gs_file    = u_gs_file
        self.d_gs_file    = d_gs_file
        self.f_gs_file    = f_gs_file
        #----------------------------------------------------------------
        save_pixels_dt_info      = cfg.read_input_option( file_unit )
        save_Q_pixels, Q_ts_file = cfg.read_output_option( file_unit )
        save_u_pixels, u_ts_file = cfg.read_output_option( file_unit )
        save_d_pixels, d_ts_file = cfg.read_output_option( file_unit )
        save_f_pixels, f_ts_file = cfg.read_output_option( file_unit )
        #----------------------------------------------------------------
        self.save_pixels_dt = save_pixels_dt_info[1]
        self.SAVE_Q_PIXELS  = save_Q_pixels
        self.SAVE_U_PIXELS  = save_u_pixels
        self.SAVE_D_PIXELS  = save_d_pixels
        self.SAVE_F_PIXELS  = save_f_pixels
        self.Q_ts_file      = Q_ts_file
        self.u_ts_file      = u_ts_file
        self.d_ts_file      = d_ts_file
        self.f_ts_file      = f_ts_file
        #----------------------------------------------------------------
        # This should also work
        #----------------------------------------------------------------        
##        save_pixels_dt_info = cfg.read_input_option( file_unit )
##        dtype_list = ['boolean', 'string']
##        save_Q_pixel_info = cfg.read_list_after_key( file_unit, dtype_list )
##        save_u_pixel_info = cfg.read_list_after_key( file_unit, dtype_list )
##        save_d_pixel_info = cfg.read_list_after_key( file_unit, dtype_list )
##        save_f_pixel_info = cfg.read_list_after_key( file_unit, dtype_list )
##        #----------------------------------------------------------------
##        self.save_pixels_dt = save_pixels_dt_info[1]
##        self.SAVE_Q_PIXELS  = save_Q_pixel_info[0]
##        self.SAVE_U_PIXELS  = save_u_pixel_info[0]
##        self.SAVE_D_PIXELS  = save_d_pixel_info[0]
##        self.SAVE_F_PIXELS  = save_f_pixel_info[0]
##        self.Q_ts_file      = save_Q_pixel_info[1]
##        self.u_ts_file      = save_u_pixel_info[1]
##        self.d_ts_file      = save_d_pixel_info[1]
##        self.f_ts_file      = save_f_pixel_info[1]
        #----------------------------------------------------------------
        
        #-----------------------
        # Close the config file
        #-----------------------
        file_unit.close()

        #---------------------------------------------------------
        # Make sure that all "save_dts" are larger or equal to
        # the specified process dt.  There is no point in saving
        # results more often than they change.
        # Issue a message to this effect if any are smaller ??
        #---------------------------------------------------------
        self.save_grid_dt   = maximum(self.save_grid_dt,   self.dt)
        self.save_pixels_dt = maximum(self.save_pixels_dt, self.dt)
        
    #   read_config_file()
    #-------------------------------------------------------------------
    def embed_child_components(self):

        #------------------------------------------------
        # Instantiate and embed "process components"
        # in the place of the CCA ports.
        #------------------------------------------------
        # But how do we choose a given process method
        # such as "kinematic wave" ??
        #------------------------------------------------
        import basins
        import met_base
        import snow_degree_day
        import ET_priestley_taylor
        import infil_green_ampt
        import GW_darcy_layers
        import diversions_fraction_method
        import ice_base

        self.bp = basins.basins_component()
        self.mp = met_base.met_component()
        self.sp = snow_degree_day.snow_component()
        self.ep = ET_priestley_taylor.ET_component()
        self.ip = infil_green_ampt.infil_component()
        self.gp = GW_darcy_layers.GW_component()
        self.dp = diversions_fraction_method.diversions_component()
        self.iip= ice_base.ice_component()   ######
        
    #   embed_child_components()
    #-------------------------------------------------------------------
    def add_child_ports(self):

        self.add_child_port('sp', 'mp')
        self.add_child_port('sp', 'bp')
        #----------------------------------
        self.add_child_port('ep', 'cp', SELF=True)
        self.add_child_port('ep', 'sp')
        self.add_child_port('ep', 'ip')
        self.add_child_port('ep', 'gp')
        self.add_child_port('ep', 'mp')
        self.add_child_port('ep', 'bp')
        #----------------------------------
        self.add_child_port('ip', 'mp')
        self.add_child_port('ip', 'sp')
        self.add_child_port('ip', 'ep')
        self.add_child_port('ip', 'gp')
        self.add_child_port('ip', 'bp')
        #----------------------------------
        self.add_child_port('gp', 'cp', SELF=True)
        self.add_child_port('gp', 'ip')
        self.add_child_port('gp', 'bp')
        #----------------------------------
        self.add_child_port('dp', 'cp', SELF=True)
        self.add_child_port('dp', 'bp')
        #----------------------------------
        self.add_child_port('mp', 'sp')
        self.add_child_port('mp', 'bp')
        ## self.add_child_port('mp', 'pp')   ##### in future ?
        #----------------------------------
        self.add_child_port('iip', 'mp')
        self.add_child_port('iip', 'sp')
        self.add_child_port('iip', 'bp')
        
    #   add_child_ports()
    #-------------------------------------------------------------------
    def initialize_ports(self):

        #-------------------------------------------------
        # Initialize the process objects/components
        # This is also where output files are opened.
        #-------------------------------------------------
        # bp must be initialized first, since other inits
        # use the outlet_ID, etc.
        #--------------------------------------------------
        DEBUG = True
        if (self.bp.get_status() != 'initialized'):        # basin vars
            self.bp.initialize(directory=self.directory,
                               data_prefix=self.data_prefix,
                               case_prefix=self.case_prefix)
            if (DEBUG): print 'CHANNELS component initialized BASINS.\n'
            
        if (self.mp.get_status() != 'initialized'):      # met vars + precip 
            self.mp.initialize(directory=self.directory,
                               data_prefix=self.data_prefix,
                               case_prefix=self.case_prefix)
            if (DEBUG): print 'CHANNELS component initialized METEOROLOGY.\n'
            
##        if (self.pp.get_status() != 'initialized'):        # precip vars
##            self.pp.initialize(directory=self.directory,
##                               data_prefix=self.data_prefix,
##                               case_prefix=self.case_prefix) 

        if (self.sp.get_status() != 'initialized'):        # snow vars         
            self.sp.initialize(directory=self.directory,
                               data_prefix=self.data_prefix,
                               case_prefix=self.case_prefix)
            if (DEBUG): print 'CHANNELS component initialized SNOW.\n'
            
        ## GW must get intialized before ET (9/25/09)
        if (self.gp.get_status() != 'initialized'):        # GW vars
            self.gp.initialize(directory=self.directory,
                               data_prefix=self.data_prefix,
                               case_prefix=self.case_prefix)
            if (DEBUG): print 'CHANNELS component initialized SATZONE.\n'
            
        if (self.ep.get_status() != 'initialized'):        # ET vars        
            self.ep.initialize(directory=self.directory,
                               data_prefix=self.data_prefix,
                               case_prefix=self.case_prefix)
            if (DEBUG): print 'CHANNELS component initialized EVAP.\n'

        if (self.dp.get_status() != 'initialized'):        # diversions
            self.dp.initialize(directory=self.directory,
                               data_prefix=self.data_prefix,
                               case_prefix=self.case_prefix)
            if (DEBUG): print 'CHANNELS component initialized DIVERSIONS.\n'
            
        if (self.ip.get_status() != 'initialized'):        # infil vars
            self.ip.initialize(directory=self.directory,
                               data_prefix=self.data_prefix,
                               case_prefix=self.case_prefix)
            if (DEBUG): print 'CHANNELS component initialized INFIL.\n'
            
        if (self.iip.get_status() != 'initialized'):        # ice vars
            self.iip.initialize(directory=self.directory,
                                data_prefix=self.data_prefix,
                                case_prefix=self.case_prefix)
            if (DEBUG): print 'CHANNELS component initialized ICE.\n'
            

    #   initialize_ports()
    #-------------------------------------------------------------------
    def initialize_d8_vars(self):

        #---------------------------------------------
        # Compute and store a variety of (static) D8
        # flow grid variables.  Embed structure into
        # the "channel_base" component.
        #---------------------------------------------
        self.d8 = d8_base.d8_component()
        self.d8.initialize(directory=self.directory,
                           data_prefix=self.data_prefix,
                           case_prefix=self.case_prefix,
                           SILENT=self.SILENT,
                           REPORT=self.REPORT)

        #-------------------------------------------      
        # We'll need this once we shift from using
        # "tf_d8_base.py" to the new "d8_base.py"
        #-------------------------------------------
        # self.d8.update(self.time, SILENT=False, REPORT=True)

    #   initialize_d8_vars()
    #-------------------------------------------------------------
    def initialize_computed_vars(self):
              
        #-----------------------------------------------
        # Convert bank angles from degrees to radians. 
        #-----------------------------------------------
        self.angles = self.angles * self.deg_to_rad  # [radians]
        
        #------------------------------------------------
        # 8/29/05.  Multiply ds by (unitless) sinuosity
        # Orig. ds is used by subsurface flow
        #------------------------------------------------
        # NB!  We should also divide slopes in S_bed by
        # the sinuosity, as now done here.
        #----------------------------------------------------
        # NB!  This saves a modified version of ds that
        #      is only used within the "channels" component.
        #      The original "ds" is stored within the
        #      topoflow model component and is used for
        #      subsurface flow, etc.
        #----------------------------------------------------
        ### self.d8.ds_chan = (self.sinu * ds)
        ### self.ds = (self.sinu * self.d8.ds)
        self.d8.ds = (self.sinu * self.d8.ds)  ### USE LESS MEMORY

        ###################################################
        ###################################################
        ### S_bed = (S_bed / self.sinu)     #*************
        self.slopes = (self.slopes / self.sinu)
        self.S_bed  = self.slopes
        ###################################################
        ###################################################
        
        #---------------------------
        # Initialize spatial grids
        #-----------------------------------------------
        # NB!  It is not a good idea to initialize the
        # water depth grid to a nonzero scalar value.
        #-----------------------------------------------
        print 'Initializing u, f, d grids...'
        self.u = zeros([self.ny, self.nx], dtype='Float64')
        self.f = zeros([self.ny, self.nx], dtype='Float64')
        self.d = zeros([self.ny, self.nx], dtype='Float64') + self.d0
        
        #-------------------------------------------
        # Make sure all slopes are valid & nonzero
        # since otherwise flow will accumulate
        #-------------------------------------------
        if (self.KINEMATIC_WAVE):    
            self.remove_bad_slopes()      #(3/8/07. Only Kin Wave case)
        
        #----------------------------------------
        # Initial volume of water in each pixel
        #----------------------------------------
        ## self.vol = zeros([self.ny, self.nx], dtype='Float64')
        A_cross  = self.d * (self.widths + (self.d * tan(self.angles)))
        self.vol = A_cross * self.d8.ds   #[m^3]

        self.initialize_outlet_values()
        self.initialize_peak_values()
        
        self.vol_R      = float64(0)  # (for mass balance)
        self.volume_out = float64(0)

##        w  = where(self.widths <= 0)
##        nw = size(w[0])   # (This is correct for 1D or 2D.)
##        if (nw > 0):
##            print 'WARNING:'
##            print 'Number of locations where width==0 =', nw
##            if (nw < 10):
##                print 'locations =', w
##            print ' '

    #   initialize_computed_vars()
    #-------------------------------------------------------------------
    def initialize_outlet_values(self):

        #----------------------------------
        # Retrieved and used by TopoFlow
        # for the stopping condition.
        #----------------------------------
        # Must be initialized somewhere
        #----------------------------------
        self.Q_outlet = float64(0)
        self.u_outlet = float64(0)
        self.d_outlet = float64(0)
        self.f_outlet = float64(0)  
        self.Q_last   = float64(0)
        
    #   initialize_outlet_values()  
    #-------------------------------------------------------------------
    def initialize_peak_values(self):

        #-------------------------
        # Initialize peak values
        #-------------------------
        self.Q_peak  = float64(0)
        self.T_peak  = float64(0)
        self.u_peak  = float64(0)
        self.Tu_peak = float64(0)  
        self.d_peak  = float64(0)
        self.Td_peak = float64(0)

    #   initialize_peak_values()
    #-------------------------------------------------------------------
    # def update_excess_rainrate(self):
    def update_R(self):

        #----------------------------------------
        # Compute the "excess rainrate", R.
        # Each term must have same units: [m/s]
        # Sum = net gain/loss rate over pixel.
        #----------------------------------------------------
        # R can be positive or negative.  If negative, then
        # water is removed from the surface at rate R until
        # surface water is consumed.
        #--------------------------------------------------------------
        # P  = precip_rate   [m/s]  (converted by read_input_data()).
        # SM = snowmelt rate [m/s]
        # GW = seep rate     [m/s]  (water_table intersects surface)
        # ET = evap rate     [m/s]
        # IN = infil rate    [m/s]
        # MR = icemelt rate  [m/s]
        #--------------------------------------------------------------        
        P  = self.get_port_data('P',  self.mp,  'METEOROLOGY')
        SM = self.get_port_data('SM', self.sp,  'SNOW')
        GW = self.get_port_data('GW', self.gp,  'SATZONE')
        ET = self.get_port_data('ET', self.ep,  'EVAP')
        IN = self.get_port_data('IN', self.ip,  'INFIL')
        MR = self.get_port_data('MR', self.iip, 'ICE')

##        ########################
##        # ONLY FOR TESTING
##        ########################
##        try:
##            H = self.get_port_data('H', self.iip) ## (ice depth   [m/s])
##        except:
##            H = float64(0)
##            print 'ERROR: Could not get "H" from ICE port.'
            
        #-------------------------------------------------------
        # (10/1/09)  This type conversion should not be needed
        # because it is done by get_port_data(), which is a
        # "private/local" vs. interface function/method.
        #-------------------------------------------------------
##        P  = numpy.float64(P)
##        SM = numpy.float64(SM)
##        GW = numpy.float64(GW)
##        ET = numpy.float64(ET)
##        IN = numpy.float64(IN)
##        MR = numpy.float64(MR)
##        H  = numpy.float64(H)

        #--------------
        # For testing
        #--------------        
##        print '(Pmin,  Pmax)  =', P.min(),  P.max()
##        print '(SMmin, SMmax) =', SM.min(), SM.max()
##        print '(GWmin, GWmax) =', GW.min(), GW.max()
##        print '(ETmin, ETmax) =', ET.min(), ET.max()
##        print '(INmin, INmax) =', IN.min(), IN.max()
##        print '(MRmin, MRmax) =', MR.min(), MR.max()
##        # print '(Hmin,  Hmax)  =', H.min(), H.max()
##        print ' '
        
        self.R = (P + SM + GW + MR) - (ET + IN)
            
    #   update_R()
    #-------------------------------------------------------------------
    def update_R_integral(self):

        #-----------------------------------------------
        # Update mass total for R, sum over all pixels
        #-----------------------------------------------   
        volume = double(self.R * self.da * self.dt)  # [m^3]
        if (size(volume) == 1):
            self.vol_R += (volume * self.rti.n_pixels)
        else:
            self.vol_R += sum(volume)

    #   update_R_integral()           
    #-------------------------------------------------------------------  
    def update_discharge(self):

        #---------------------------------------------------------
        # The discharge grid, Q, gives the flux of water _out_
        # of each grid cell.  This entire amount then flows
        # into one of the 8 neighbor grid cells, as indicated
        # by the D8 flow code. The update_flow_volume() function
        # is called right after this one in update() and uses
        # the Q grid.
        #---------------------------------------------------------
        # 7/15/05.  The cross-sectional area of a trapezoid is
        # given by:    Ac = d * (w + (d * tan(theta))),
        # where w is the bottom width.  If we were to
        # use: Ac = w * d, then we'd have Ac=0 when w=0.
        # We also need angle units to be radians.
        #---------------------------------------------------------

        #-----------------------------
        # Compute the discharge grid
        #-----------------------------
        L2 = self.d * tan(self.angles)
        Ac = self.d * (self.widths + L2)
        self.Q = float64(self.u * Ac)
            
        #--------------
        # For testing
        #--------------
        # print 'dmin, dmax =', self.d.min(), self.d.max()
        # print 'umin, umax =', self.u.min(), self.u.max()
        # print 'Qmin, Qmax =', self.Q.min(), self.Q.max()
        # print ' '        
        # print 'u(outlet) =', self.u[self.outlet_ID]
        # print 'Q(outlet) =', self.Q[self.outlet_ID]  ########
          
        #----------------------------------------------------
        # Wherever depth is less than z0, assume that water
        # is not flowing and set u and Q to zero.
        # However, we also need (d gt 0) to avoid a divide
        # by zero problem, even when numerators are zero.
        #----------------------------------------------------
        # FLOWING = (d > (z0/aval))
        #*** FLOWING[self.d8.noflow_IDs] = False    ;******
        # u = (u * FLOWING)
        # Q = (Q * FLOWING)
        # d = maximum(d, 0.0)    ;(allow depths lt z0, if gt 0.)

    #   update_discharge()
    #-------------------------------------------------------------------
    def update_diversions(self):

        #-----------------------------------------------------------
        # Notes: This function was written on 2/16/10 to allow a
        #        Channels component to handle sources, sinks and
        #        canals directly.
        #
        #        This function must be called between calls to
        #        update_discharge() and update_flow_volume().
        #
        #        In the current approach, the Channels component
        #        only retrieves a small amount of data from the
        #        Diversions component.  The old method required
        #        the Diversion component to retrieve the entire
        #        Q or vol grid from the Channel component at each
        #        timestep, alter it, and then pass it back.
        #-----------------------------------------------------------
        # NB!    If sources, sinks or canals use large Q-values
        #        (perhaps for testing), then it is necessary to
        #        reduce the timestep, dt, to prevent instability
        #        and oscillations.
        #-----------------------------------------------------------
        # NB!    For now, the dt used for all diversions must be
        #        the same as that of the Channel component.
        #-----------------------------------------------------------        
        # Notes: Q   = surface discharge  [m^3/s]
        #        R   = excess precip. rate  [m/s]
        #        da  = pixel area  [m^2]
        #        dt  = channel flow timestep  [s]
        #        vol = total volume of water in pixel [m^3]
        #        v2  = temp version of vol
        #        w1  = IDs of pixels that...
        #        p1  = IDs of parent pixels that...
        #-----------------------------------------------------------
        dt = self.dt  # [seconds]

        #---------------------------------
        # Add contributions from sources
        #------------------------------------------------
        # If a grid cell contains a "source", then an
        # additional Q will flow *into* that grid cell.
        #------------------------------------------------
        if (self.use_sources):
##            print 'Updating Q_sources in CHANNELS...'
##            print 'dp.has_variable("Q_sources") =', self.dp.has_variable('Q_sources')
##            print 'dp.is_scalar("Q_sources")    =', self.dp.is_scalar('Q_sources')
##            print 'dp.is_vector("Q_sources")    =', self.dp.is_vector('Q_sources')
##            print 'dp.is_grid("Q_sources")      =', self.dp.is_grid('Q_sources')
            
            Q_sources = self.dp.get_vector_double('Q_sources')
            self.vol.flat[ self.source_IDs ] += (dt * Q_sources)
            #---------------------------------------------------
            # NB! Do not add the next line!  It is incorrect
            #     because the Q grid tells how much is leaving
            #     each grid cell and Q_sources is an amount
            #     that is coming into a grid cell.
            #---------------------------------------------------
            ## self.Q.flat[ self.source_IDs] += Q_sources
            
        #------------------------------------
        # Subtract contributions from sinks
        #-------------------------------------------------------
        # NB! If a grid cell contains a "sink" or the upstream
        #     end of a "canal", then the discharge *leaving*
        #     that grid cell must be reduced accordingly.
        #     This changes the Q grid and must be done before
        #     using the Q grid for further calculations below.
        #-------------------------------------------------------
        # NB! self.vol MUST be updated here.
        #-------------------------------------------------------  
        if (self.use_sinks):
            ## print 'Updating Q_sinks in CHANNELS...'
            Q_sinks = self.dp.get_vector_double('Q_sinks')
            self.Q.flat[ self.sink_IDs ]   -= Q_sinks
            self.vol.flat[ self.sink_IDs ] -= (Q_sinks * self.dt)  ######
            
        #-----------------------------------------
        # Add/subtract contributions from canals
        #-------------------------------------------------------
        # NB! This changes the Q grid and must be done before
        #     using the Q grid for further calculations below.
        #-------------------------------------------------------
        # NB! self.vol MUST be updated here.
        #-------------------------------------------------------        
        if (self.use_canals):
            # print 'Updating Q_canals_in in CHANNELS...'
            Q_canals_in = (self.canal_Q_fractions * self.Q.flat[ self.canal_in_IDs ])
            self.Q.flat[ self.canal_in_IDs ]   -= Q_canals_in           
            self.vol.flat[ self.canal_in_IDs ] -= (dt * Q_canals_in)
            
            #--------------------------------------------------------------
            # Now put this amount into the upstream end of the canal.
            # Diversions component will advance it toward downstream end.
            #--------------------------------------------------------------
            self.dp.set_vector_double('Q_canals_in', Q_canals_in)

            #-------------------------------------------
            # Add flow volume that enters from a canal
            #-------------------------------------------
            ## print 'Updating Q_canals_out CHANNELS...'
            Q_canals_out = self.dp.get_vector_double('Q_canals_out')
            self.vol.flat[ self.canal_out_IDs ] += (dt * Q_canals_out)

        #---------------------------------------------
        # I think this could now be moved back out
        # to the TopoFlow driver. (2/16/10)
        #---------------------------------------------
        # It is placed after update_discharge() and
        # update_flow_volume() have both been called.
        #---------------------------------------------
        self.dp.update( self.time_sec )
        ############################################
        ############################################        
        ## print 'Updating DIVERSIONS...'
        ## print ' '
        
    #   update_diversions()
    #-------------------------------------------------------------------
    def update_flow_volume(self):

        #-----------------------------------------------------------
        # Notes: This function must be called after
        #        update_discharge() and update_diversions().
        #-----------------------------------------------------------        
        # Notes: Q   = surface discharge  [m^3/s]
        #        R   = excess precip. rate  [m/s]
        #        da  = pixel area  [m^2]
        #        dt  = channel flow timestep  [s]
        #        vol = total volume of water in pixel [m^3]
        #        v2  = temp version of vol
        #        w1  = IDs of pixels that...
        #        p1  = IDs of parent pixels that...
        #-----------------------------------------------------------
        dt = self.dt  # [seconds]

        #----------------------------------------------------
        # Add contribution (or loss ?) from excess rainrate
        #----------------------------------------------------
        self.vol += (self.R * self.da) * dt
    
        #-----------------------------------------
        # Add contributions from neighbor pixels
        #-------------------------------------------------------------
        # Each grid cell passes flow to *one* downstream neighbor.
        # Note that multiple grid cells can flow toward a given grid
        # cell, so a grid cell ID may occur in d8.p1 and d8.p2, etc.
        #-------------------------------------------------------------
        # (2/16/10)  RETEST THIS.  Before, a copy called "v2" was
        # used but this doesn't seem to be necessary.
        #-------------------------------------------------------------        
        if (self.d8.p1_OK):    
            self.vol[ self.d8.p1 ] += (dt * self.Q[self.d8.w1])
        if (self.d8.p2_OK):    
            self.vol[ self.d8.p2 ] += (dt * self.Q[self.d8.w2])
        if (self.d8.p3_OK):    
            self.vol[ self.d8.p3 ] += (dt * self.Q[self.d8.w3])
        if (self.d8.p4_OK):    
            self.vol[ self.d8.p4 ] += (dt * self.Q[self.d8.w4])
        if (self.d8.p5_OK):    
            self.vol[ self.d8.p5 ] += (dt * self.Q[self.d8.w5])
        if (self.d8.p6_OK):    
            self.vol[ self.d8.p6 ] += (dt * self.Q[self.d8.w6])
        if (self.d8.p7_OK):    
            self.vol[ self.d8.p7 ] += (dt * self.Q[self.d8.w7])
        if (self.d8.p8_OK):    
            self.vol[ self.d8.p8 ] += (dt * self.Q[self.d8.w8])

        #----------------------------------------------------
        # Subtract the amount that flows out to D8 neighbor
        #----------------------------------------------------
        self.vol -= (self.Q * dt)
   
        #--------------------------------------------------------
        # While R can be positive or negative, the surface flow
        # volume must always be nonnegative. This also ensures
        # that the flow depth is nonnegative.  (7/13/06)
        #--------------------------------------------------------
        self.vol = numpy.maximum(self.vol, 0.0)
              
    #   update_flow_volume
    #-------------------------------------------------------------------
    def update_flow_depth(self):

        #-----------------------------------------------------------
        # Notes: 7/18/05.  Modified to use the equation for volume
        #        of a trapezoidal channel:  vol = Ac * ds, where
        #        Ac=d*[w + d*tan(t)], and to solve the resulting
        #        quadratic (discarding neg. root) for new depth, d.

        #        8/29/05.  Now original ds is used for subsurface
        #        flow and there is a ds_chan which can include a
        #        sinuosity greater than 1.  This may be especially
        #        important for larger pixel sizes.


        #        Removed (ds > 1) here which was only meant to
        #        avoid a "divide by zero" error at pixels where
        #        (ds eq 0).  This isn't necessary since the
        #        Flow_Lengths function in utils_TF.pro never
        #        returns a value of zero.
        #----------------------------------------------------------      
        self.d_last = self.d.copy()  #####
        d = self.d   # (local alias)
        
        #--------------------------------------------
        # Are bank angles given as scalar or grid ?
        #--------------------------------------------
        na = size(self.angles)
        if (na == 1):    
            #------------------------------------
            # Bank angles are given as a scalar
            #------------------------------------
            if (self.angles == 0.0):    
                d = self.vol / (self.widths * self.d8.ds)
            else:    
                #-----------------------------------------------
                # See Notes for TF_Tan function in tf_utils.py
                #-----------------------------------------------
                # term1 = float32(2.0) * TF_Tan(self.angles)
                term1 = float32(2.0) * tan(self.angles)
                arg   =  float32(2.0) * term1 * self.vol / self.d8.ds
                arg  += self.widths ** float32(2.0)
                d     = (sqrt(arg) - self.widths) / term1
        else:
            #----------------------------------
            # Bank angles are given as a grid
            #----------------------------------
            wz   = where(self.angles == 0)
            nwz  = size(wz[0])
            wzc  = where(self.angles != 0)
            nwzc = size(wzc[0])
            
            if (nwz != 0):
                A_top = self.widths[wz] * self.d8.ds[wz]
                ## A_top = self.widths[wz] * self.d8.ds_chan[wz]            
                d[wz] = self.vol[wz] / A_top
            
            if (nwzc != 0):    
                #-----------------------------------------------
                # See Notes for TF_Tan function in tf_utils.py
                #-----------------------------------------------
                ## term1  = float32(2.0) * TF_Tan(self.angles[wzc])
                term1  = float32(2.0) * tan(self.angles[wzc])
                arg    =  float32(2.0) * term1 * self.vol[wzc] / self.d8.ds[wzc]
                arg   += self.widths[wzc] ** float32(2.0)
                d[wzc] = (sqrt(arg) - self.widths[wzc]) / term1

        #------------------------------------------
        # Set depth values on edges to zero since
        # they become spikes (no outflow) 7/15/06
        #------------------------------------------
        d[self.d8.noflow_IDs] = 0.0   #####  later use d.flat ?? #####

        #------------------------------------------------
        # 4/19/06.  Force flow depth to be positive ?
        #------------------------------------------------
        # This seems to be needed with the non-Richards
        # infiltration routines when starting with zero
        # depth everywhere, since all water infiltrates
        # for some period of time.  It also seems to be
        # needed more for short rainfall records to
        # avoid a negative flow depth error.
        #------------------------------------------------
        # 7/13/06.  Still needed for Richards method
        #------------------------------------------------
        self.d = maximum(d, 0.0)
        
    #   update_flow_depth
    #-------------------------------------------------------------------
    def update_free_surface_slope(self):

        #-----------------------------------------------------------
        # Notes:  It is assumed that the flow directions don't
        #         change even though the free surface is changing.
        #-----------------------------------------------------------
        delta_d     = (self.d - self.d[self.d8.parent_IDs])
        self.S_free = self.S_bed + (delta_d / self.d8.ds)
        
        #--------------------------------------------
        # Don't do this; negative slopes are needed
        # to decelerate flow in dynamic wave case
        # and for backwater effects.
        #--------------------------------------------
        # Set negative slopes to zero
        #------------------------------
        ###  self.S_free = maximum(self.S_free, 0)

    #   update_free_surface_slope()
    #-------------------------------------------------------------------
    def update_trapezoid_Rh(self):

        #-------------------------------------------------------------
        # Notes: Compute the hydraulic radius of a trapezoid that:
        #          (1) has a bed width of wb >= 0 (0 for triangular)
        #          (2) has a bank angle of theta (0 for rectangular)
        #          (3) is filled with water to a depth of d.
        #        The units of wb and d are meters.  The units of
        #        theta are assumed to be degrees and are converted.
        #-------------------------------------------------------------
        # NB!    wb should never be zero, so PW can never be 0,
        #        which would produce a NaN (divide by zero).
        #-------------------------------------------------------------
        #        See Notes for TF_Tan function in utils_TF.pro
        #            AW = d * (wb + (d * TF_Tan(theta_rad)) )
        #-------------------------------------------------------------
        d  = self.d        # (local synonyms)
        wb = self.widths   # (trapezoid bottom width)
        
        theta_rad = (self.angles * self.deg_to_rad)

##        print 'dmin, dmax =', d.min(),  d.max()
##        print 'wmin, wmax =', wb.min(), wb.max()
##        print 'amin, amax =', theta_rad.min(), theta_rad.max()
        
        #---------------------------------------------------------
        # Compute hydraulic radius grid for trapezoidal channels
        #---------------------------------------------------------        
        AW = d * (wb + (d * tan(theta_rad)) )      
        PW = wb + (float32(2.0) * d / cos(theta_rad) )
        Rh = (AW / PW)

        #------------------------------------
        # Force edge pixels to have Rh = 0.
        # This will make u = 0 there also.
        #------------------------------------
        Rh[self.d8.noflow_IDs] = float64(0)        
##        w  = where(wb <= 0)
##        nw = size(w[0])
##        if (nw > 0): Rh[w] = float64(0)
        
        self.Rh = Rh

    #   update_trapezoid_Rh()
    #-------------------------------------------------------------------
    def update_velocity(self):

        #---------------------------------------------------------
        # Note: Do nothing now unless this method is overridden
        #       by a particular method of computing velocity.
        #---------------------------------------------------------
        print "Warning: update_velocity() method is inactive."
        
        # print 'KINEMATIC WAVE =', self.KINEMATIC_WAVE
        # print 'DIFFUSIVE WAVE =', self.DIFFUSIVE_WAVE
        # print 'DYNAMIC WAVE   =', self.DYNAMIC_WAVE

    #   update_velocity()
    #-------------------------------------------------------------------
    def update_velocity_on_edges(self):

        #---------------------------------
        # Force edge pixels to have u=0.
        #----------------------------------------
        # Large slope around 1 flows into small
        # slope & leads to a negative velocity.
        #----------------------------------------
        self.u[self.d8.noflow_IDs] = float64(0)
        
    #   update_velocity_on_edges()
    #-------------------------------------------------------------
    def update_outlet_values(self):
        
        #---------------------------------
        # Save computed values at outlet
        #   (used by TopoFlow)
        #---------------------------------
        self.Q_outlet = self.Q[self.outlet_ID]
        self.u_outlet = self.u[self.outlet_ID]
        self.d_outlet = self.d[self.outlet_ID]
        self.f_outlet = self.f[self.outlet_ID]
        
##        self.Q_outlet = self.Q.flat[self.outlet_ID]
##        self.u_outlet = self.u.flat[self.outlet_ID]
##        self.d_outlet = self.d.flat[self.outlet_ID]
##        self.f_outlet = self.f.flat[self.outlet_ID]
        
    #   update_outlet_values()
    #-------------------------------------------------------------
    def update_peak_values(self):
    
        if (self.Q_outlet > self.Q_peak):    
            self.Q_peak  = self.Q_outlet
            self.T_peak  = self.time_min      # (time to peak)
        #-----------------------------------
        if (self.u_outlet > self.u_peak):
            self.u_peak  = self.u_outlet
            self.Tu_peak = self.time_min
        #-----------------------------------
        if (self.d_outlet > self.d_peak):    
            self.d_peak  = self.d_outlet
            self.Td_peak = self.time_min

    #   update_peak_values()
    #-------------------------------------------------------------
    def update_volume_out(self):

        #----------------------------------------------------
        # Rename "volume_out" to "vol_Q" for consistency ??
        #----------------------------------------------------
        self.volume_out += (self.Q_outlet * self.dt)
        ## self.volume_out += (self.Q[self.outlet_ID] * self.dt)
        
    #   update_volume_out()
    #-------------------------------------------------------------
    def update_mins_and_maxes(self, REPORT=False):

        self.Q_min = self.Q.min()
        self.Q_max = self.Q.max()
        self.u_min = self.u.min()
        self.u_max = self.u.max()
        self.d_min = self.d.min()
        self.d_max = self.d.max()

        if (REPORT):
            print '(dmin, dmax) =', self.d_min, self.d_max
            print '(umin, umax) =', self.u_min, self.u_max
            print '(Qmin, Qmax) =', self.Q_min, self.Q_max
            print ' '
            
    #   update_mins_and_maxes()
    #-------------------------------------------------------------------
    def check_flow_depth(self):

        OK = True
        d  = self.d
        dt = self.dt
        nx = self.nx   #################
        
        #---------------------------------
        # All all flow depths positive ?
        #---------------------------------
        wbad = where(logical_or( d < 0.0, logical_not(isfinite(d)) ))
        nbad = size(wbad[0])       
        if (nbad == 0):    
            return OK

        OK = False
        dmin = d[wbad].min()
        star_line = '*******************************************'
        
        msg = [ star_line, \
               'ERROR: Simulation aborted.', ' ', \
               'Negative depth found: ' + str(dmin), \
               'Time step may be too large.', \
               'Time step:      ' + str(dt) + ' [s]', ' ']
        for k in xrange(len(msg)):
            print msg[k]
        
        #-------------------------------------------
        # If not too many, print actual velocities
        #-------------------------------------------
        if (nbad < 30):          
            brow = wbad[0][0]
            bcol = wbad[1][0]
##            badi = wbad[0]
##            bcol = (badi % nx)
##            brow = (badi / nx)
            crstr = str(bcol) + ', ' + str(brow)

            msg = ['(Column, Row):  ' + crstr, \
                   'Flow depth:     ' + str(d[brow, bcol])]
            for k in xrange(len(msg)):
                print msg[k]

        print star_line 
        print ' '
        return OK

    #   check_flow_depth
    #-------------------------------------------------------------------
    def check_flow_velocity(self):

        OK = True
        u  = self.u
        dt = self.dt
        nx = self.nx
        
        #--------------------------------
        # Are all velocities positive ?
        #--------------------------------
        wbad = where(logical_or( u < 0.0, logical_not(isfinite(u)) ))
        nbad = size(wbad[0])
        if (nbad == 0):    
            return OK

        OK = False
        umin = u[wbad].min()
        star_line = '*******************************************'
        msg = [ star_line, \
               'ERROR: Simulation aborted.', ' ', \
               'Negative or NaN velocity found: ' + str(umin), \
               'Time step may be too large.', \
               'Time step:      ' + str(dt) + ' [s]', ' ']
        for k in xrange(len(msg)):
            print msg[k]

        #-------------------------------------------
        # If not too many, print actual velocities
        #-------------------------------------------
        if (nbad < 30):
            brow = wbad[0][0]
            bcol = wbad[1][0]
##            badi = wbad[0]
##            bcol = (badi % nx)
##            brow = (badi / nx)
            crstr = str(bcol) + ', ' + str(brow)

            msg = ['(Column, Row):  ' + crstr, \
                   'Velocity:       ' + str(u[brow, bcol])]
            for k in xrange(len(msg)):
                print msg[k]

        print star_line
        print ' '
        return OK

            
##        umin = u[wbad].min()
##        badi = wbad[0]
##        bcol = (badi % nx)
##        brow = (badi / nx)
##        crstr = str(bcol) + ', ' + str(brow)
##        msg = array([' ', \
##                     '*******************************************', \
##                     'ERROR: Simulation aborted.', ' ', \
##                     'Negative velocity found: ' + str(umin), \
##                     'Time step may be too large.', ' ', \
##                     '(Column, Row):  ' + crstr, \
##                     'Velocity:       ' + str(u[badi]), \
##                     'Time step:      ' + str(dt) + ' [s]', \
##                     '*******************************************', ' '])
##        for k in xrange(size(msg)):
##            print msg[k]

##        return OK                          


    #   check_flow_velocity
    #-------------------------------------------------------------------  
    def open_input_files(self):

        #self.code_unit = model_input.open_file(self.code_type,  self.code_file)
        self.slope_unit = model_input.open_file(self.slope_type, self.slope_file)
        if (self.MANNING):
            self.nval_unit  = model_input.open_file(self.nval_type,  self.nval_file)
        if (self.LAW_OF_WALL):
            self.z0val_unit = model_input.open_file(self.z0val_type, self.z0val_file)
        self.width_unit = model_input.open_file(self.width_type, self.width_file)
        self.angle_unit = model_input.open_file(self.angle_type, self.angle_file)
        self.sinu_unit  = model_input.open_file(self.sinu_type,  self.sinu_file)
        self.d0_unit    = model_input.open_file(self.d0_type,    self.d0_file)

    #   open_input_files()        
    #-------------------------------------------------------------------  
    def read_input_files(self):

        #---------------------------------------------------
        # The flow codes are always a grid, size of DEM.
        #---------------------------------------------------
        # NB! model_input.py also has a read_grid() function.
        #---------------------------------------------------        
        rti = self.rti
        print 'Reading D8 flow grid (in TF)...'
        self.codes = rtg_files.read_grid(self.code_file, rti,
                                         RTG_type='BYTE')
        print ' '
        
        #-------------------------------------------------------
        # All grids are assumed to have a data type of Float32.
        #-------------------------------------------------------
        slopes = model_input.read_next(self.slope_unit, self.slope_type, rti)
        if (slopes != None): self.slopes = slopes
        
        # If EOF was reached, hopefully numpy's "fromfile"
        # returns None, so that the stored value will be
        # the last value that was read.

        if (self.MANNING):
            nvals = model_input.read_next(self.nval_unit, self.nval_type, rti)
            if (nvals != None):
                self.nvals     = nvals
                self.nvals_min = nvals.min()
                self.nvals_max = nvals.max()
                
        if (self.LAW_OF_WALL):
            z0vals = model_input.read_next(self.z0val_unit, self.z0val_type, rti)
            if (z0vals != None):
                self.z0vals     = z0vals
                self.z0vals_min = z0vals.min()
                self.z0vals_max = z0vals.max()
        
        widths = model_input.read_next(self.width_unit, self.width_type, rti)
        if (widths != None): self.widths = widths
        
        angles = model_input.read_next(self.angle_unit, self.angle_type, rti)
        if (angles != None): self.angles = angles

        sinu = model_input.read_next(self.sinu_unit, self.sinu_type, rti)
        if (sinu != None): self.sinu = sinu
        
        d0 = model_input.read_next(self.d0_unit, self.d0_type, rti)
        if (d0 != None): self.d0 = d0

        ## codes = model_input.read_grid(self.code_unit, \
        ##                            self.code_type, rti, dtype='UInt8')
        ## if (codes != None): self.codes = codes

    #   read_input_files()     
    #-------------------------------------------------------------------  
    def close_input_files(self):

        # if not(self.slope_unit.closed):
        # if (self.slope_unit != None):

        #-------------------------------------------------
        # NB!  self.code_unit was never defined as read.
        #-------------------------------------------------
        # if (self.code_file  != ''): self.code_unit.close()
        
        if (self.slope_file != ''): self.slope_unit.close()
        if (self.nval_file  != ''): self.nval_unit.close()
        if (self.z0val_file != ''): self.z0val_unit.close()
        if (self.width_file != ''): self.width_unit.close()
        if (self.angle_file != ''): self.angle_unit.close()
        if (self.sinu_file  != ''): self.sinu_unit.close()
        if (self.d0_file    != ''): self.d0_unit.close()

    #   close_input_files()       
    #-------------------------------------------------------------------  
    def update_outfile_names(self):

        #--------------------------------------------------------
        # Notes:  Whenever the case_prefix changes (e.g. user
        #         changes it and clicks the Next button in the
        #         "run var" panel), all output filenames should
        #         be reset to new defaults.
        #--------------------------------------------------------
        self.Q_gs_file = (self.case_prefix + '_2D-Q.rts')
        self.u_gs_file = (self.case_prefix + '_2D-u.rts')
        self.d_gs_file = (self.case_prefix + '_2D-d.rts')
        self.f_gs_file = (self.case_prefix + '_2D-f.rts')
        #---------------------------------------------------
        self.Q_ts_file = (self.case_prefix + '_0D-Q.txt')
        self.u_ts_file = (self.case_prefix + '_0D-u.txt')
        self.d_ts_file = (self.case_prefix + '_0D-d.txt')
        self.f_ts_file = (self.case_prefix + '_0D-f.txt')

    #   update_outfile_names()     
    #-------------------------------------------------------------------  
    def open_output_files(self):
         
        #--------------------------------------
        # Open new files to write grid stacks
        #--------------------------------------
        if (self.SAVE_Q_GRIDS):   
            self.Q_gs_unit = open_new_gs_file(self.Q_gs_file, self.rti,
                                              grid_name='Q',
                                              long_name='volumetric_discharge',
                                              units_name='m^3/s')
            if (self.Q_gs_unit == -1):
                self.SAVE_Q_GRIDS = False
            
        if (self.SAVE_U_GRIDS):    
            self.u_gs_unit = open_new_gs_file(self.u_gs_file, self.rti,
                                              grid_name='u',
                                              long_name='mean_channel_flow_velocity',
                                              units_name='m/s')
        
        if (self.SAVE_D_GRIDS):    
            self.d_gs_unit = open_new_gs_file(self.d_gs_file, self.rti,
                                              grid_name='d',
                                              long_name='max_channel_flow_depth',
                                              units_name='m')

        if (self.SAVE_F_GRIDS):    
            self.f_gs_unit = open_new_gs_file(self.f_gs_file, self.rti,
                                              grid_name='f',
                                              long_name='friction_factor',
                                              units_name='none')

        #---------------------------------------
        # Open text files to write time series
        #---------------------------------------
        IDs = self.outlet_IDs
        if (self.SAVE_Q_PIXELS):    
            self.Q_ts_unit = open_new_ts_file( self.Q_ts_file, 'Q', IDs)
            
        if (self.SAVE_U_PIXELS):
            self.u_ts_unit = open_new_ts_file( self.u_ts_file, 'u', IDs)
            
        if (self.SAVE_D_PIXELS):    
            self.d_ts_unit = open_new_ts_file( self.d_ts_file, 'd', IDs)
            
        if (self.SAVE_F_PIXELS):    
            self.f_ts_unit = open_new_ts_file( self.f_ts_file, 'f', IDs)

    #   open_output_files()
    #-------------------------------------------------------------------  
    def write_output_files(self, time_seconds=None):

        #---------------------------------------------------------
        # Notes:  This function was written to use only model
        #         time (maybe from a caller) in seconds, and
        #         the save_grid_dt and save_pixels_dt parameters
        #         read by read_config_file().
        #
        #         read_config_file() makes sure that all of
        #         the "save_dts" are larger than or equal to the
        #         process dt.
        #---------------------------------------------------------
        
        #-----------------------------------------
        # Allows time to be passed from a caller
        #-----------------------------------------
        if (time_seconds is None):
            time_seconds = self.time_sec
        model_time = int(time_seconds)
        
        #----------------------------------------
        # Save computed values at sampled times
        #----------------------------------------
        if (model_time % int(self.save_grid_dt) == 0):
            self.save_grids()
        if (model_time % int(self.save_pixels_dt) == 0):
            self.save_pixel_values()

        #----------------------------------------
        # Save computed values at sampled times
        #----------------------------------------
##        if ((self.time_index % self.grid_save_step) == 0):
##             self.save_grids()
##        if ((self.time_index % self.pixel_save_step) == 0):
##             self.save_pixel_values()
        
    #   write_output_files()
    #-------------------------------------------------------------------  
    def close_output_files(self):

        if (self.SAVE_Q_GRIDS):  self.Q_gs_unit.close()   
        if (self.SAVE_U_GRIDS):  self.u_gs_unit.close()    
        if (self.SAVE_D_GRIDS):  self.d_gs_unit.close()   
        if (self.SAVE_F_GRIDS):  self.f_gs_unit.close()
        #---------------------------------------------------
        if (self.SAVE_Q_PIXELS): self.Q_ts_unit.close()   
        if (self.SAVE_U_PIXELS): self.u_ts_unit.close()    
        if (self.SAVE_D_PIXELS): self.d_ts_unit.close()    
        if (self.SAVE_F_PIXELS): self.f_ts_unit.close()

    #   close_output_files()              
    #-------------------------------------------------------------------  
    def save_grids(self):
        
        #---------------------------------------------------
        # Notes:  Each variable is saved as a grid whether
        #         it is a scalar or already a (2D) grid.
        #---------------------------------------------------
        nx = self.nx
        ny = self.ny

        if (self.SAVE_Q_GRIDS):
            save_as_grid_to_file(self.Q_gs_unit, self.Q, 'Q', nx, ny)
            
        if (self.SAVE_U_GRIDS):
            save_as_grid_to_file(self.u_gs_unit, self.u, 'u', nx, ny)
            
        if (self.SAVE_D_GRIDS):
            save_as_grid_to_file(self.d_gs_unit, self.d, 'd', nx, ny)

        if (self.SAVE_F_GRIDS):
            save_as_grid_to_file(self.f_gs_unit, self.f, 'f', nx, ny)    

    #   save_grids()
    #-------------------------------------------------------------------  
    def save_pixel_values(self):   ##### save_time_series_data(self)  #######
        
        IDs      = self.outlet_IDs
        time_min = self.time_min
        
        if (self.SAVE_Q_PIXELS):
            write_ts_file_line(self.Q_ts_unit, time_min, self.Q, IDs)
                    
        if (self.SAVE_U_PIXELS):
            write_ts_file_line(self.u_ts_unit, time_min, self.u, IDs)
            
        if (self.SAVE_D_PIXELS):
            write_ts_file_line(self.d_ts_unit, time_min, self.d, IDs)
            
        if (self.SAVE_F_PIXELS):
            write_ts_file_line(self.f_ts_unit, time_min, self.f, IDs)

    #   save_pixel_values()
    #-------------------------------------------------------------------
    def manning_formula(self, S):

        #---------------------------------------------------------
        # Notes: R = (A/P) = hydraulic radius [m]
        #        N = Manning's roughness coefficient
        #            (usually in the range 0.012 to 0.035)
        #        S = bed slope (assumed equal to friction slope)

        #        R,S, and N may be 2D arrays.

        #        If length units are all *feet*, then an extra
        #        factor of 1.49 must be applied.  If units are
        #        meters, no such factor is needed.

        #        Note that Q = Ac * u, where Ac is cross-section
        #        area.  For a trapezoid, Ac does not equal w*d.
        #---------------------------------------------------------
        ##  if (N == None): N = float32(0.03)
        
        u = (self.Rh ** self.two_thirds) * sqrt(S) / self.nvals
        
        #------------------------------
        # Add a hydraulic jump option
        # for when u gets too big ??
        #------------------------------
        
        #--------------------------------------
        # Option to return friction factor, f
        # (Correct this for Rh vs. d.)
        # **** See update_velocity_dynamic()
        #--------------------------------------
        #*** self.f = (self.g * self.d * self.S) / (u^2d)
        
        return u
    
    #   manning_formula()
    #-------------------------------------------------------------------
    def law_of_the_wall(self, S):

        #---------------------------------------------------------
        # Notes: u  = flow velocity  [m/s]
        #        d  = flow depth [m]
        #        z0 = roughness height
        #        S  = bed slope (assumed equal to friction slope)

        #        g     = 9.81 = gravitation constant [m/s^2]
        #        kappa = 0.41 = von Karman's constant
        #        aval  = 0.48 = integration constant

        #        sqrt(g)/kappa = 7.6393d
        #        smoothness = (aval / z0) * d
        #        f = (kappa / alog(smoothness))^2d
        #        tau_bed = rho_w * f * u^2 = rho_w * g * d * S

        #        d, S, and z0 can be arrays.

        #        To make default z0 correspond to default
        #        Manning's n, can use this approximation:
        #        z0 = a * (2.34 * sqrt(9.81) * n / kappa)^6d
        #        For n=0.03, this gives: z0 = 0.011417
        #        However, for n=0.3, it gives: z0 = 11417.413
        #        which is 11.4 km!  So the approximation only
        #        holds within some range of values.
        #--------------------------------------------------------
##        if (self.z0vals == None):    
##            self.z0vals = float64(0.011417)   # (about 1 cm)

        smoothness = (self.aval / self.z0vals) * self.d
          
        #-----------------------------
        # Make sure (smoothness > 1)
        #-----------------------------
        smoothness = maximum(smoothness, float64(1.1))

        u = self.law_const * sqrt(self.Rh * S) * log(smoothness)
                   
        ## d = self.d
        ## u = float64(7.676) * sqrt(d * S) * log(smoothness)
        
        #------------------------------
        # Add a hydraulic jump option
        # for when u gets too big ??
        #------------------------------
        
        #--------------------------------------
        # Option to return friction factor, f
        # **** See update_velocity_dynamic()
        #--------------------------------------
        ## self.f = (self.kappa / alog(smoothness))^2d
        
        return u
    
    #   law_of_the_wall()
    #-------------------------------------------------------------------
    def print_status_report(self): 

        #----------------------------------------------------
        # Wherever depth is less than z0, assume that water
        # is not flowing and set u and Q to zero.
        # However, we also need (d gt 0) to avoid a divide
        # by zero problem, even when numerators are zero.
        #----------------------------------------------------
        # FLOWING = (d > (z0/aval))
        #*** FLOWING[noflow_IDs] = False    ;******
        
        wflow    = where(FLOWING != 0)
        n_flow   = size(wflow[0])
        n_pixels = self.rti.n_pixels
        percent  = float32(100.0) * (float32(n_flow) / n_pixels)
        fstr = ('%5.1f' % percent) + '%'
        # fstr = idl_func.string(percent, format='(F5.1)').strip() + '%'
        print ' Percentage of pixels with flow = ' + fstr
        print ' '

        self.update_mins_and_maxes(REPORT=True)
 
        wmax  = where(self.Q == self.Q_max)
        nwmax = size(wmax[0])
        print ' Max(Q) occurs at: ' + str(wmax[0])
        #print,' Max attained at ', nwmax, ' pixels.'
        print ' '
        print '-------------------------------------------------'

    #   print_status_report()         
    #-------------------------------------------------------------------
    def remove_bad_slopes(self, FLOAT=False):

        #------------------------------------------------------------
        # Notes: The main purpose of this routine is to find
        #        pixels that have nonpositive slopes and replace
        #        then with the smallest value that occurs anywhere
        #        in the input slope grid.  For example, pixels on
        #        the edges of the DEM will have a slope of zero.

        #        With the Kinematic Wave option, flow cannot leave
        #        a pixel that has a slope of zero and the depth
        #        increases in an unrealistic manner to create a
        #        spike in the depth grid.

        #        It would be better, of course, if there were
        #        no zero-slope pixels in the DEM.  We could use
        #        an "Imposed gradient DEM" to get slopes or some
        #        method of "profile smoothing".

        #        It is possible for the flow code to be nonzero
        #        at a pixel that has NaN for its slope. For these
        #        pixels, we also set the slope to our min value.

        #        7/18/05. Broke this out into separate procedure.
        #------------------------------------------------------------

        #-----------------------------------
        # Are there any "bad" pixels ?
        # If not, return with no messages.
        #-----------------------------------  
        wb = where(logical_or((self.slopes <= 0.0), \
                              logical_not(isfinite(self.slopes))))
        nbad = size(wb[0])
        print 'size(slopes) =', size(self.slopes)
        print 'size(wb)     =', nbad
        
        wg = where(invert(logical_or((self.slopes <= 0.0), \
                                     logical_not(isfinite(self.slopes)))))
        ngood = size(wg[0])
        if (nbad == 0) or (ngood == 0):
            return
        
        #---------------------------------------------
        # Find smallest positive value in slope grid
        # and replace the "bad" values with smin.
        #---------------------------------------------
        print '-------------------------------------------------'
        print 'WARNING: Zero or negative slopes found.'
        print '         Replacing them with smallest slope.'
        print '         Use "Profile smoothing tool" instead.'
        S_min = self.slopes[wg].min()
        S_max = self.slopes[wg].max()
        print '         min(S) = ' + str(S_min)
        print '         max(S) = ' + str(S_max)
        print '-------------------------------------------------'
        print ' '
        self.slopes[wb] = S_min
        
        #--------------------------------
        # Convert data type to double ?
        #--------------------------------
        if (FLOAT):    
            self.slopes = float32(self.slopes)
        else:    
            self.slopes = float64(self.slopes)
        
    #   remove_bad_slopes
    #-------------------------------------------------------------------

#-------------------------------------------------------------------
def Trapezoid_Rh(d, wb, theta):

    #-------------------------------------------------------------
    # Notes: Compute the hydraulic radius of a trapezoid that:
    #          (1) has a bed width of wb >= 0 (0 for triangular)
    #          (2) has a bank angle of theta (0 for rectangular)
    #          (3) is filled with water to a depth of d.
    #        The units of wb and d are meters.  The units of
    #        theta are assumed to be degrees and are converted.
    #-------------------------------------------------------------
    # NB!    wb should never be zero, so PW can never be 0,
    #        which would produce a NaN (divide by zero).
    #-------------------------------------------------------------
    #        See Notes for TF_Tan function in utils_TF.pro
    #            AW = d * (wb + (d * TF_Tan(theta_rad)) )
    #-------------------------------------------------------------    
    theta_rad = (theta * numpy.pi / 180.0)
    
    AW = d * (wb + (d * tan(theta_rad)) )      
    PW = wb + (float32(2.0) * d / cos(theta_rad) )
    Rh = (AW / PW)

    w  = where(wb <= 0)
    nw = size(w[0])
    
    return Rh

#   Trapezoid_Rh()
#-------------------------------------------------------------------
def Manning_Formula(Rh, S, nvals):

    #---------------------------------------------------------
    # Notes: R = (A/P) = hydraulic radius [m]
    #        N = Manning's roughness coefficient
    #            (usually in the range 0.012 to 0.035)
    #        S = bed slope (assumed equal to friction slope)

    #        R,S, and N may be 2D arrays.

    #        If length units are all *feet*, then an extra
    #        factor of 1.49 must be applied.  If units are
    #        meters, no such factor is needed.

    #        Note that Q = Ac * u, where Ac is cross-section
    #        area.  For a trapezoid, Ac does not equal w*d.
    #---------------------------------------------------------
    ##  if (N == None): N = float32(0.03)

    two_thirds = float64(2.0) / 3.0
    
    u = (Rh ** two_thirds) * sqrt(S) / nvals
    
    #------------------------------
    # Add a hydraulic jump option
    # for when u gets too big ??
    #------------------------------
    
    return u

#   Manning_Formula()
#-------------------------------------------------------------------
def Law_of_the_Wall(d, Rh, S, z0vals):

    #---------------------------------------------------------
    # Notes: u  = flow velocity  [m/s]
    #        d  = flow depth [m]
    #        z0 = roughness height
    #        S  = bed slope (assumed equal to friction slope)

    #        g     = 9.81 = gravitation constant [m/s^2]
    #        kappa = 0.41 = von Karman's constant
    #        aval  = 0.48 = integration constant

    #        sqrt(g)/kappa = 7.6393d
    #        smoothness = (aval / z0) * d
    #        f = (kappa / alog(smoothness))^2d
    #        tau_bed = rho_w * f * u^2 = rho_w * g * d * S

    #        d, S, and z0 can be arrays.

    #        To make default z0 correspond to default
    #        Manning's n, can use this approximation:
    #        z0 = a * (2.34 * sqrt(9.81) * n / kappa)^6d
    #        For n=0.03, this gives: z0 = 0.011417
    #        However, for n=0.3, it gives: z0 = 11417.413
    #        which is 11.4 km!  So the approximation only
    #        holds within some range of values.
    #--------------------------------------------------------
##        if (self.z0vals == None):    
##            self.z0vals = float64(0.011417)   # (about 1 cm)

    #------------------------
    # Define some constants
    #------------------------
    g          = float64(9.81)    # (gravitation const.)
    aval       = float64(0.476)   # (integration const.)
    kappa      = float64(0.408)   # (von Karman's const.)
    law_const  = sqrt(g) / kappa
        
    smoothness = (aval / z0vals) * d
      
    #-----------------------------
    # Make sure (smoothness > 1)
    #-----------------------------
    smoothness = maximum(smoothness, float64(1.1))

    u = law_const * sqrt(Rh * S) * log(smoothness)
    
    #------------------------------
    # Add a hydraulic jump option
    # for when u gets too big ??
    #------------------------------
    
    return u

#   Law_of_the_Wall()
#-------------------------------------------------------------------                 
