
## Copyright (c) 2001-2010, Scott D. Peckham
## January, May, July 2009
## May 2010 (changes to unit_test() and read_cfg_file()

#-----------------------------------------------------------------------
#  NOTES:  This file defines a "diffusive wave" channel flow component
#          and related functions.  It inherits from the channels
#          "base class" in "channels_base.py".
#-----------------------------------------------------------------------
#
#  unit_test()
#
#  class channels_component
#
#      get_attribute()     # (10/26/11)
#      update_velocity()
#
#-----------------------------------------------------------------------

from numpy import *
import numpy

import cfg_files as cfg
import channels_base
import tf_utils

#-----------------------------------------------------------------------
def unit_test():

    c = channels_component()
    c.CCA   = False
    c.DEBUG = False
    
    #-----------------------------------------
    # This function adjusts for the platform
    # and can be changed in "tf_utils.py".
    #-----------------------------------------
    cfg_directory = tf_utils.TF_Test_Directory()
    cfg_prefix    = tf_utils.TF_Test_Case_Prefix()
    c.site_prefix = 'Treynor'  ###########      

    #-------------------------------------------
    # Call initialize() and call update() once
    #-------------------------------------------
##    print 'STATUS =', c.get_status()
##    c.initialize(mode="driver")
##    print 'STATUS =', c.get_status()
##    time_sec = float64(0)
##    c.update(time_sec)
##    print 'STATUS =', c.get_status()

    c.run_model( cfg_directory=cfg_directory,
                 cfg_prefix=cfg_prefix )
    
##    print 'KINEMATIC WAVE =', c.KINEMATIC_WAVE
##    print 'DIFFUSIVE WAVE =', c.DIFFUSIVE_WAVE
##    print 'DYNAMIC WAVE   =', c.DYNAMIC_WAVE
##    print ' '
##    print 'nx          =', c.nx
##    print 'ny          =', c.ny
    
#   unit_test()
#-----------------------------------------------------------------------
class channels_component(channels_base.channels_component):

    #-------------------------------------------------------------------
    def get_attribute(self, att_name):

        #-----------------------------------------------------------
        # This is done in channels_base.set_computed_input_vars()
        #-----------------------------------------------------------
        # self.KINEMATIC_WAVE = False 
        # self.DIFFUSIVE_WAVE = True 
        # self.DYNAMIC_WAVE   = False

        map = {'comp_name':          'ChannelsDiffWave',
               'version':            '3.1',
               'model_name':         'Channels_Diffusive_Wave',
               'model_family':       'TopoFlow',
               'cfg_template_file':  'Channels_Diffusive_Wave.cfg.in',
               'cfg_extension':      '_channels_diffusive_wave.cfg',
               'cmt_var_prefix':     '/ChannelsDiffWave/Input/Var/',
               'gui_xml_file':       '/home/csdms/cca/topoflow/3.1/src/share/cmt/gui/Channels_Diffusive_Wave.xml',
               'dialog_title':       'Channels: Diffusive Wave Parameters',
               'time_step_type':     'fixed',
               'time_units':         'seconds',
               'mesh_type':          'uniform',
               'author_name':        'Scott Peckham'}

        try:
            return map[ att_name.lower() ]
        except:
            print '###################################################'
            print ' ERROR: Could not find attribute: ' + att_name
            print '###################################################'
            print ' '

    #   get_attribute() 
    #-------------------------------------------------------------------
    def update_velocity(self):

        #---------------------------------------------------------
        # Notes: Compute u from d and S_bed.  (7/13/05 version)

        #        cv        = channel vars structure
        #        cv.nval   = Manning's n values (grid)
        #        cv.z0_val = z0 roughness values (grid)
        #        cv.width  = channel bottom widths (grid)
        #        cv.angle  = channel bank angles (grid)

        #        Could use slopes in cv also, but S_bed has
        #        been modified from those values to impose a
        #        minimum slope that is nonzero.

        #        Rh = hydraulic radius (trapezoid here)

        #        S = S_bed  for KINEMATIC_WAVE option.
        #        S = S_free for DIFFUSIVE_WAVE_option.
        #---------------------------------------------------------

        #----------------------------
        # Update free surface slope
        #----------------------------
        self.update_free_surface_slope()   # (called S_free)

        #-------------------------------------------
        # Compute velocity using S_free vs. S_bed
        #-------------------------------------------
        # NB! This involves computing sqrt(S_free)
        # so disallow "backflow" this way ?
        #-------------------------------------------
        self.S_free = maximum(self.S_free, 0.0)        ############
        
        #------------------------
        # Use Manning's formula
        #------------------------
        if (self.MANNING):    
            self.u = self.manning_formula(self.S_free)
        
        #--------------------------------------
        # Use the Logarithmic Law of the Wall
        #--------------------------------------
        if (self.LAW_OF_WALL):    
            self.u = self.law_of_the_wall(self.S_free)

        #----------------------------------------
        # Allow negative velocity (backflow) ??
        # But to which child pixel ??
        #----------------------------------------
        # wn  = where(S_free < 0)
        # nwn = size(wn[0])
        # self.S_free = abs(self.S_free)
        # if (nwn > 0): self.u[wn] = -1.0 * self.u[wn]
        
    #    update_velocity()                       
    #-------------------------------------------------------------------


         
