
## This routine still uses "self.method" but is not used now.
            
## Copyright (c) 2001-2010, Scott D. Peckham
## January 2009  (converted from IDL)
## April, August 2009
## May 2010 (changes to initialize() and read_cfg_file()

#-----------------------------------------------------------------------
# Notes:  Carefully recheck "update_snow_vars()" function, which
#         must get and set values via ports.

#         New update() function not tested yet !!!

#         When we run out of data in input files, we
#         continue to use the last value read, by default.
#         For rainrate, however, we want to drop to 0, as
#         is done here.

#         Unusual in use of set_port_data() in CSDMS_base.py.
#-----------------------------------------------------------------------
#
#  unit_test()
#
#  class precip_component   (inherits from CSDMS_base.py)
#
#      set_constants()
#      get_input_items()      # (not written yet)
#      get_output_items()     # (not written yet)
#      -----------------------
#      initialize()
#      update()
#      finalize()
#      -----------------------
#      get_cca_ports()           # (in CSDMS_base.py)
#      embed_child_components()
#      add_child_ports()
#      initialize_ports()
#      release_cca_ports()       # (in CSDMS_base.py)
#      ------------------------    
#      update_precip_rate()
#      update_P_max()
#      update_P_integral()
#      update_snow_vars()
#      update_method1_vars()
#      update_duration_sum()
#      ------------------------
#      open_input_files()
#      read_input_files()
#      close_input_files()

#-----------------------------------------------------------------------   

from numpy import *
import numpy

# import os

import CSDMS_base
import model_input
import tf_utils

#-----------------------------------------------------------------------
def unit_test():

    c = precip_component()
    c.CCA   = False
    c.DEBUG = False
    
    #-----------------------------------------
    # This function adjusts for the platform
    # and can be changed in "tf_utils.py".
    #-----------------------------------------
    cfg_directory = tf_utils.TF_Test_Directory()
    cfg_prefix    = tf_utils.TF_Test_Case_Prefix()
    c.site_prefix = 'Treynor'  #############    

    #-------------------------------------------
    # Call initialize() and call update() once
    #-------------------------------------------
##    print 'STATUS =', c.get_status()
##    c.initialize( mode="driver" )
##    print 'STATUS =', c.get_status()
##    time_sec = float64(0)
##    c.update(time_sec)
##    print 'STATUS =', c.get_status()

    c.run_model( cfg_directory=cfg_directory,
                 cfg_prefix=cfg_prefix )
    
#   unit_test()
#-----------------------------------------------------------------------
class precip_component(CSDMS_base.CSDMS_component):

    #-------------------------------------------------------------------
    def set_constants(self):

        #-----------------------------------
        # Constants not changeable by user
        #-----------------------------------
        self.mmph_to_mps = (float64(1) / float64(3600000))
        self.forever     = float64(999999999)  # [minutes]

        #------------------------------------------------
        # Only needed for method 1, where all rates and
        # durations are read as 1D arrays from GUI.
        # Method 1 may be removed in a future version.
        #------------------------------------------------
        self.method1_rates     = None
        self.method1_durations = None
        self.method1_n_rates   = 0
        
   #    set_constants()                
    #-------------------------------------------------------------------
    def initialize(self, cfg_prefix=None, mode="nondriver",
                   SILENT=False):

        #---------------------------------------------------------
        # NB! Rainrate units are assumed to be stored with units
        #     of [mm/hr], but are converted to [m/s] prior to
        #     storage in the data structure by Read_Var, via the
        #     new FACTOR keyword.  (1/19/07)
        #---------------------------------------------------------
        #  Any variable that must be reset at the beginning of
        #  every run or may have changed between runs should be
        #  put in initialize() vs. __init__().  This includes
        #  counters like "method1_index", etc.
        #---------------------------------------------------------
        if not(SILENT):
           print 'Precipitation component: Initializing...'
           
        self.status     = 'initializing'  # (OpenMI 2.0 convention)
        self.mode       = mode
        self.cfg_prefix = cfg_prefix
        
        #-----------------------------------------------
        # Load component parameters from a config file
        # Will use cfg_file from above.
        #-----------------------------------------------
        self.set_constants()
        self.initialize_config_vars() 
        self.read_grid_info()
        self.initialize_basin_vars()  # (5/14/10)
        
        #----------------------------------------
        # Establish "P" as a synonym for "rate",
        # which should persist throughout.
        #----------------------------------------
        self.P = self.rate
        
        #-------------------------------
        # Return if no method selected
        #-------------------------------
        if (self.comp_status == 'Disabled'):
            self.P = float64(0)
            self.DONE = True
            self.status = 'initialized'       
            return
            
        #-----------------------------------------------
        # Read from files as needed to initialize vars 
        #-----------------------------------------------
        self.open_input_files()
        self.read_input_files()

        #----------------------------------------
        # Initialize all time-related variables
        #----------------------------------------
        self.initialize_time_vars()
        
        #--------------------------------------------------
        # Initialize duration_sum with the first duration
        #--------------------------------------------------
        self.duration_sum = float64(0)  # (see update())
        ## self.duration_sum = self.duration

        #------------------------------------------
        # Initialize max_rate with the first rate
        #------------------------------------------------
        # Note: Need this here because rate may be
        #       zero at the end of update_precip_rate()
        #------------------------------------------------
        self.max_rate = self.rate
        self.vol_P    = float64(0)  # (for mass balance)
        
        #---------------------------
        # Only needed for Method 1
        #---------------------------
        self.method1_index = 0  # (here vs. __init__)
           
        #------------------------------------
        # Are all input variables scalars ?
        #------------------------------------
        types = array([self.rate_type, self.duration_type])
        self.all_scalars = (types.max() < 2)

        self.initialize_required_components(mode)
               
##        self.open_output_files()  # (not needed yet)
        
        self.status = 'initialized'
        
    #   initialize()
    #-------------------------------------------------------------------
    def update(self, time_seconds=None):

        #---------------------------------------------------------
        # NB! (8/20/09) Modified to use fixed dt vs. "durations".
        #     Additional changes may need to be made elsewhere.
        #---------------------------------------------------------        
        # NB!  Precip is currently handled differently from the
        # other processes, because variable durations are allowed.
        # This means the Precipitation function must be called
        # for every time step to see if it has changed.  Note
        # that n is not an argument to this function.
        #---------------------------------------------------------
        # Notes: mp = meteorology port
        #        sp = snow port
        #
        #        P  = precip rate with units of [m/s],
        #             converted from [mm/hr] by the
        #             routines in GUI_precip.pro
        #        Will this work if some intervals have
        #        precip. rates of zero ??
        #--------------------------------------------------------

        #--------------------------------------------------
        # Note: self.P already set to 0 by initialize()
        #--------------------------------------------------
        if (self.comp_status == 'Disabled'): return
        self.status = 'updating'  # (OpenMI)

        #-------------------------
        # Update computed values 
        #-------------------------
        self.update_precip_rate()
        self.update_P_max()
        self.update_P_integral()
        self.update_snow_vars()   # (snow depth due to snowfall)

        #------------------------
        # Update internal clock
        #------------------------
        self.update_duration_sum()  # (RECHECK HOW INITIALIZED)
        self.update_time()
        
        #-------------------------------------------
        # Read next precip vars from input files ?
        #-------------------------------------------
        # Note that read_input_files() is called
        # by initialize() and these values must be
        # used for "update" calls before reading
        # new ones.
        #-------------------------------------------
        ## if (time_min > self.duration_sum):    # (PREVIOUSLY)
        if (self.method != 1):
            self.read_input_files()
        else:
            #-----------------------------------
            # The way Method 1 works is unique
            # to the Precip process.
            #-----------------------------------
            self.update_method1_vars()
            
##        #-----------------------------
##        # Is it time to update yet ?
##        #-----------------------------
##        if (time_min > self.duration_sum):
##            #-----------------------------------------
##            # Read next precip vars from input files ?
##            #-----------------------------------------
##            if (self.method != 1):
##                self.read_input_files()
##            else:
##                #-----------------------------------
##                # The way Method 1 works is unique
##                # to the Precip process.
##                #-----------------------------------
##                self.update_method1_vars()
##
##            #-------------------------
##            # Update computed values 
##            #-------------------------
##            self.update_duration_sum()
##            self.update_precip_rate()
##            self.update_P_integral()
##
##        self.update_snow_vars()

        self.status = 'updated'  # (OpenMI)
        
    #   update()
    #-------------------------------------------------------------------
    def finalize(self):

        self.status = 'finalizing'  # (OpenMI) 
        self.close_input_files()
        ## self.close_output_files()
        self.status = 'finalized'  # (OpenMI)
        print 'Precipitation component: Finished.'

        #---------------------------
        # Release all of the ports
        #----------------------------------------
        # Make this call in "finalize()" method
        # of the component's CCA Imple file
        #----------------------------------------
        # self.release_cca_ports( port_names, d_services )
        
    #   finalize()
    #-------------------------------------------------------------------
    def embed_child_components(self):

        #------------------------------------------------
        # Note: Don't call this if (self.CCA == True).
        #------------------------------------------------
        # Instantiate and embed "process components"
        # in the place of the CCA ports.
        #------------------------------------------------
        # But how do we choose a given process method
        # such as "kinematic wave" ??
        #------------------------------------------------
        import met_base
        import snow_base
        
        self.mp = met_base.met_component()
        self.sp = snow_base.snow_component()

    #   embed_child_components()
    #-------------------------------------------------------------------
    def add_child_ports(self):

        print 'Warning: "add_child_ports() is not implemented yet.'
        
    #   add_child_ports()       
    #-------------------------------------------------------------------
    def initialize_ports(self):
        
        #-----------------------------------------------
        # Initialize the process objects/components
        # This is also where output files are opened.
        #-----------------------------------------------        
        if (self.mp.get_status() != 'initialized'):        # met vars 
            self.mp.initialize( cfg_prefix=self.cfg_prefix )
            
        if (self.sp.get_status() != 'initialized'):        # snow vars         
            self.sp.initialize( cfg_prefix=self.cfg_prefix )

    #   initialize_ports()        
    #-------------------------------------------------------------------
    def update_precip_rate(self):

        #------------------------------------------------------
        # Note:  If we run out of rainrate data while reading
        #        from files, read_input_files() will set the
        #        rainrate to zero.
        #------------------------------------------------------
        if (self.method == 0):    
            self.rate = float32(0.0)    # (no precip)
        elif (self.method == 1):
            #-----------------------------------------------
            # Read next scalar rate from saved list
            # update() calls update_method1_vars to do this.
            #-----------------------------------------------
            pass                                  
        elif (self.method == 2):
            #----------------------------------------
            # Use rate that we just read from input
            # file, which may be a scalar or grid
            #----------------------------------------
            pass
        else:
            raise RuntimeError('No match found for expression')

        #-------------------------------------------------
        # NB!  P is just another reference to self.rate.
        #-------------------------------------------------
        # If (self.method == 1), self.rate gets updated
        # when update() calls update_method1_vars().
        #-------------------------------------------------
        # If (self.method == 2), self.rate gets updated
        # by read_input_files().
        #-------------------------------------------------
        self.P = self.rate

    #   update_precip_rate()
    #-------------------------------------------------------------------
    def update_P_max(self):

        #-----------------------------------------
        # Save the maximum precip. rate in [m/s]
        #-----------------------------------------
        self.max_rate = maximum(self.max_rate, self.rate.max())
        
    #   update_P_max()
    #-------------------------------------------------------------------
    def update_P_integral(self):

        #---------------------------------------------------
        # Notes: This can be used for mass balance checks,
        #        such as now done by update_mass_totals()
        #        in topoflow.py.  The "dt" here should be
        #        TopoFlow's "main dt" vs. the process dt.
        
        #        dV[i] = P[i] * da[i] * dt, dV = sum(dV[i])
        #---------------------------------------------------
        
        #------------------------------------------------
        # Update mass total for P, sum over all pixels
        #------------------------------------------------   
        volume = double(self.P * self.da * self.dt)  # [m^3]
        if (size(volume) == 1):
            self.vol_P += (volume * self.n_pixels)
        else:
            self.vol_P += sum(volume)
        
    #   update_P_integral()
    #-------------------------------------------------------------------
    def update_snow_vars(self):

        #---------------------------------------------
        # If P or T_air is a grid, then we must have
        # h_swe and h_snow be grids.  This is set
        # up at start of Route_Flow.
        #---------------------------------------------
        T_air    = self.get_port_data('T_air',  self.mp)
        h_swe    = self.get_port_data('h_swe',  self.sp)
        h_snow   = self.get_port_data('h_snow', self.sp)
        #--------------------------------------------------        
        rho_H2O  = self.sp.get_scalar_double('rho_H2O')
        rho_snow = self.sp.get_scalar_double('rho_snow')
        ### main_dt  = self.cp.get_scalar_double('dt')
        
        #----------------------------------------------
        # Update snow water equivalent and snow depth
        #----------------------------------------------
        # (3/14/07) New method that works regardless
        # of whether P and T are scalars or grids.
        #---------------------------------------------
        # (8/20/09) Precip process now has constant
        # timestep, dt, which should be used here.
        #---------------------------------------------        
        dh_swe  = (self.P * (T_air <= 0)) * self.dt
        h_swe  += dh_swe
        ratio   = (rho_H2O / rho_snow)
        h_snow  = h_swe * ratio

        #----------------------------------------------------
        # Use "set_values" call to update h_snow and h_swe?
        #----------------------------------------------------
        self.set_port_data('h_snow', h_snow, self.sp)
        self.set_port_data('h_swe',  h_swe,  self.sp)
        
        #---------------------------------------------
        # Where precip falls as snow and gets saved
        # to SWE, P must be set to zero so it can't
        # go on to produce runoff.
        #-----------------------------------------------
        # A special situation occurs if P is a scalar
        # and T is a grid.  P must then be converted
        # to a grid in order to be able to reflect the
        # fact that water was stored as snow in some
        # pixels and not at others.  This happens
        # automatically as written here. (3/14/07)
        # If P and T are both scalars, P stays scalar.
        ################################################
        # NB!  Should still work in Python, because:
        #   (1.5 * True = 1.5, 1.5 * False = 0)
        ################################################
        # NB!  This could be slowing things down a lot
        #      because it happens every timestep, not
        #      just when precip rate is updated.  Is
        #      there some way to test out of this ??
        ################################################
        ################################################
        self.P = (self.P * (dh_swe == 0))

        #--------------
        # For testing
        #--------------
        # nT = size(mv.T_air)
        # nP = size(self.P)
        # if (nT == 1):
        #     tstr = str(mv.T_air)
        #     print '   T_air = ' + tstr
        # if (nP == 1):
        #     pstr = str(self.P)
        #     print '   P     = ' + pstr
        
        #-------------------------------------
        # Message about rainfall or snowfall
        #-------------------------------------
        # Note that rain and snow may fall
        # simultaneously at different pixels
        #-------------------------------------
        # Reporting this slows model too much
        # but can be used for testing
        #-------------------------------------
        # w1 = where(dh_swe > 0)
        # n1 = size(w1[0])
        # if (n1 > 0): print '   Snow is falling...'
        #--------------------------------------------------
        # dh_rain = (P * (mv.T_air > 0))
        # w2 = where(dh_rain > 0)
        # n2 = size(w2[0])
        # if (n2 > 0): print '   Rain is falling...'

    #   update_snow_vars()               
    #-------------------------------------------------------------------
    def update_duration_sum(self):

        #--------------------------------------------------------
        # Note: duration_sum could also be called "update_time"
        #       which might make the code more intuitive.
        #--------------------------------------------------------
        if (self.duration != None):
            self.duration_sum += self.duration
        else:
            self.duration_sum = self.forever  ########

    #   update_duration_sum() 
    #-------------------------------------------------------------------
    def update_method1_vars(self):
        
        #-----------------------------------
        # Read next vars from saved arrays
        # that were read directly from GUI
        #-----------------------------------
        self.method1_index += 1
        m1_index = self.method1_index
        
        if (m1_index < self.method1_n_rates):
            self.rate     = self.method1_rates[m1_index]
            self.duration = self.method1_durations[m1_index]
        else:
            self.rate     = float32(0)
            self.duration = self.forever

    #   update_method1_vars()
    #-------------------------------------------------------------------  
    def open_input_files(self):

        # For testing
##        import os
##        print 'self.rate_type     =', self.rate_type
##        print 'self.duration_type =', self.duration_type
##        print 'self.rate_file     =', self.rate_file
##        print 'self.duration_file =', self.duration_file
##        print 'directory          =', os.getcwd()
##        print ' '
        
        self.rate_unit      = model_input.open_file(self.rate_type, \
                                            self.rate_file)
        self.duration_unit  = model_input.open_file(self.duration_type,  \
                                            self.duration_file)

    #   open_input_files()   
    #-------------------------------------------------------------------
    def read_input_files(self):

        #-------------------------------------------------------
        # All grids are assumed to have a data type of Float32.
        #-------------------------------------------------------
        rate = model_input.read_next(self.rate_unit, self.rate_type, \
                                  self.rti, factor=self.mmph_to_mps)
        if (rate != None):
            self.rate = rate
        else:
            #--------------------------------
            # Out of data, set rate to zero
            #--------------------------------
            self.rate = float32(0)
        
        duration = model_input.read_next(self.duration_unit, \
                                      self.duration_type, self.rti)
        if (duration != None):
            self.duration = duration
        else:
            self.duration = self.forever    ##################
            
        #-------------------
        # For testing only
        #-------------------
        if (rate != None):
            if (numpy.size(rate) == 1):
                print 'rate_file =', self.rate_file
                print 'rate      =', rate, ' [m/s]'
                print 'duration  =', duration, '[min]'
                print ' '

    #   read_input_files()        
    #-------------------------------------------------------------------  
    def close_input_files(self):

        if (self.rate_file     != ''):  self.rate_unit.close()
        if (self.duration_file != ''):  self.duration_unit.close()

    #   close_input_files()   
    #-------------------------------------------------------------------
        
