
## Copyright (c) 2001-2010, Scott D. Peckham
## January 2009  # (converted from IDL)
## April, May, July, August 2009
## Feb 2010 (channels comp. now calls dp.update() itself.)
## May 2010 (changes to initialize() and read_cfg_file()

###############################################################
#  Note: We could have the channels component make most of
#        the process update() calls.  This may clean up the
#        appearance in the CMT.

#  NB! T Some of the  "print_*" functions are not ready yet.

#      In CCA mode, all update() method calls require the
#      "time_sec" argument.
###############################################################

#--------------------------------------------------------------------
#  Notes:  This is an object-oriented, Python version of the
#          TopoFlow spatial, hydrologic model, originally developed
#          in IDL.  It now attempts to follow the OpenMI interface
#          standard (in a CCA context) for every process.
#--------------------------------------------------------------------

#--------------------------------------------------------------------
#  Notes:  Later go through code line-by-line to find places where
#          numpy calculations (e.g. where) could be done more
#          efficiently.
#--------------------------------------------------------------------
#
#  unit_test()
#
#  class topoflow      (inherits from CSDMS_base.py)
#
#      __init__()             ###################### DO WE NEED THIS NOW  ??? (10/26/11)
#      --------------------
#      get_attribute()         # (10/26/11)
#      set_constants()
#      run_model()
#      initialize()
#      update()
#      finalize()
#      ---------------------
#      get_cca_ports()           # (in CSDMS_base.py)
#      get_cca_port_info()       # (5/11/10)
#      embed_child_components()
#      add_child_ports()
#      initialize_ports()
#      release_cca_ports()       # (in CSDMS_base.py)
#      ---------------------
#      compute_update_steps()    ########
#      -----------------------------------
#      check_finished()
#      check_steady_state()   ###
#      check_interrupt()
#      -----------------------------------
#      initialize_stop_vars()
#      initialize_mass_totals()       # (OBSOLETE ??)
#      initialize_GUI()               # (commented out)
#      initialize_hydrograph_plot()   # (commented out)
#      -----------------------------------
#      update_mass_totals()           # (OBSOLETE ??)
#      update_hydrograph_plot()
#      -----------------------------------
#      print_final_report()
#      print_mins_and_maxes()
#      print_uniform_precip_data()
#      print_dimless_number_data()
#      print_mass_balance_report()

#-----------------------------------------------------------------------
from numpy import *
import numpy

import os
import time

import cfg_files as cfg
import CSDMS_base

from tf_utils import TF_Print, TF_String
import tf_utils  # (must come after the "from" line ???)

## from model_output import *

#-----------------------------------------------------------------------
def unit_test():

    c = topoflow()
    c.CCA   = False
    c.DEBUG = False
    
    #-----------------------------------------
    # This function adjusts for the platform
    # and can be changed in "tf_utils.py".
    #-----------------------------------------
    cfg_directory   = tf_utils.TF_Test_Directory()
    cfg_prefix      = tf_utils.TF_Test_Case_Prefix()
    c.site_prefix = 'Treynor'  #############    

    #-------------------------------------------
    # Call initialize() and call update() once
    #-------------------------------------------
##    print 'STATUS =', c.get_status()
##    c.initialize( mode="driver" )
##    print 'STATUS =', c.get_status()
##    time_sec = float64(0)
##    c.update(time_sec)
##    print 'STATUS =', c.get_status()

    c.run_model( cfg_directory=cfg_directory,
                 cfg_prefix=cfg_prefix )

#   unit_test()
#-----------------------------------------------------------------------
class topoflow(CSDMS_base.CSDMS_component):
    
    def __init__(self, VERBOSE=False, WRITE_LOG=True, \
                 PLOT=False, USE_GUI=False):

        CSDMS_base.CSDMS_component.__init__(self)
##        self.CCA     = tf_utils.TF_Use_CCA()
##        self.DEBUG   = False
##        self.SILENT  = True
##        self.REPORT  = False
        
        self.OK      = True
        self.comment = None
  
        self.WRITE_LOG    = WRITE_LOG
        self.VERBOSE      = VERBOSE
        self.PLOT         = PLOT

        self.status = 'created'    # (OpenMI)
        
    #   __init__()
    #-------------------------------------------------------------------
    def get_attribute(self, att_name):

        map = {'comp_name':          'TopoFlow',
               'version':            '3.1',
               'model_name':         'TopoFlow_Driver',
               'model_family':       'TopoFlow',
               'cfg_template_file':  'TopoFlow.cfg.in',
               'cfg_extension':      '_topoflow.cfg',
               'cmt_var_prefix':     '/TopoFlow/Input/Var/',
               'gui_xml_file':       '/home/csdms/cca/topoflow/3.1/src/share/cmt/gui/TopoFlow.xml',
               'dialog_title':       'TopoFlow: Driver Parameters',
               'time_step_type':     'fixed',
               'time_units':         'seconds',
               'mesh_type':          'none',
               'author_name':        'Scott Peckham'}

        try:
            return map[ att_name.lower() ]
        except:
            print '###################################################'
            print ' ERROR: Could not find attribute: ' + att_name
            print '###################################################'
            print ' '

    #   get_attribute() 
    #-------------------------------------------------------------------
    def set_constants(self):

        #------------------------
        # Define some constants
        #------------------------
        self.mps_to_mmph = float64(3600000)   # [m/s] to [mm/hr]
        
    #   set_constants()       
    #-------------------------------------------------------------------
    def initialize(self, cfg_prefix=None, mode="nondriver",
                   SILENT=False):

        #------------------------------------------------------
        # Note:  If using as a CCA component, then we need to
        #        call "get_cca_ports()" before calling this
        #        "initialize()" method in the component's CCA
        #        Impl file.
        #------------------------------------------------------
        if not(SILENT):
            print 'TopoFlow component: Initializing...'
        
        self.status     = 'initializing'  # (OpenMI 2.0 convention)
        self.mode       = mode
        self.cfg_prefix = cfg_prefix
        
        #-----------------------------------------------
        # Load component parameters from a config file
        #-----------------------------------------------
        self.set_constants()
        self.initialize_config_vars() 
        self.read_grid_info()
        self.initialize_basin_vars()  # (5/14/10)

        #----------------------------------
        # Has component been turned off ?
        #----------------------------------
        if (self.comp_status == 'Disabled'):
            if not(SILENT):
                print 'TopoFlow component: Disabled.'
            self.DONE = True
            self.status = 'initialized.'  # (OpenMI 2.0 convention)
            return
        
        dc = (self.out_directory + self.case_prefix)
        self.comment_file = dc + '_comments.txt'
        self.log_file     = dc + '.log'

        #############################################################        
        #############################################################
        # This will be done by PortQueue in the future. (10/26/11)
        # Correct "mode" must be used when CMT.initialize() calls
        # BMI.initialize().
        #############################################################
        #############################################################
        self.initialize_required_components(mode)
    
        #-----------------------
        # Initialize variables
        #-----------------------
        self.initialize_time_vars()  # (uses cp.dt from above)
        self.initialize_stop_vars()   #### (do this with CFG file ?)
        self.Q_outlet = float64(0)
        
        #-------------------------
        # Compute "update steps"
        #--------------------------------------------
        # Process modules must be initialized first
        #--------------------------------------------
        self.compute_update_steps()
                          
        #### self.nz = self.ip.get_scalar_long('nz')   #######
        
        #------------------------------------
        # Check if output options are valid
        #------------------------------------
##        outlet_IDs, n_outlets, outlet_ID, OK = \
##            Check_Output_Options(grid_vars, rv, cv, sv, ev, iv, gv, nx, ny)
      
        #---------------------
        # Open the logfile ?      *** (Append or overwrite ??)
        #---------------------
        if (self.WRITE_LOG):
            if (self.log_file == None):
                self.log_file = (self.case_prefix + '_LOG.txt')
            print 'Opening log file:'
            print '    log_file =', self.log_file
            self.log_unit = open(self.log_file, 'w')

        #----------------------
        # Print start message
        #----------------------
        TF_Print(' ')
        TF_Print('Starting TopoFlow model run...')
        hline = ''.ljust(60, '-')
        TF_Print(hline)
        self.status = 'initialized'
        
    #   initialize()        
    #-------------------------------------------------------------            
    def update(self, time_sec):

        #-------------------------------
        # Check for interrupt by user ?
        #-------------------------------
        # OK = self.check_interrupt()
        # if not(OK):
        #    self.status = 'stopped'
        #    return

        # self.DEBUG = True  ########################
        
        self.status = 'updating'
        OK = True
        ## if (self.VERBOSE):
        if (self.mode == 'driver'):
            self.print_time_and_value(self.Q_outlet, 'Q_out', '[m^3/s]')
        ## self.update_hydrograph_plot()

        ## print '##### In topoflow.update(), time_sec =', self.time_sec
        
        #------------------------------------------------
        # Update channel process first, using the first
        # values that were read for each process.
        #------------------------------------------------
        self.cp.update(self.time_sec)  # (uses others, so after them)
        if (self.DEBUG): print 'TF UPDATED Channels Component...'

        OK = (self.cp.get_status() != 'failed')
        if not(OK):
            if (self.DEBUG): print 'CHANNELS component failed.'
            self.Q_outlet = float64(0)   # (why is this here again?)
            self.DONE = True
            return

        if (self.DEBUG): print 'MADE IT PAST fail check...'
        
        #------------------------------------
        # Step each process forward in time
        #------------------------------------
        UPDATE_MP = (self.time_index % self.mp_update_step) == 0
        UPDATE_SP = (self.time_index % self.sp_update_step) == 0
        UPDATE_EP = (self.time_index % self.ep_update_step) == 0
        UPDATE_IP = (self.time_index % self.ip_update_step) == 0
        UPDATE_GP = (self.time_index % self.gp_update_step) == 0
        UPDATE_IIP= (self.time_index % self.iip_update_step)== 0
        ## UPDATE_DP = (self.time_index % self.dp_update_step) == 0

        if (self.DEBUG): print 'MADE IT PAST update checks...'
        
        # NB! Precip process now uses fixed dt. (August 09)
        # NB! Precip has been absorbed into Met (Sept 09)

        if (UPDATE_MP):
            self.mp.update(self.time_sec)    # update met_vars
            if (self.DEBUG): print 'TF UPDATED Met Component...'
        if (UPDATE_SP):
            self.sp.update(self.time_sec)    # update snow_vars
            if (self.DEBUG): print 'TF UPDATED Snow Component...'
        if (UPDATE_EP):
            self.ep.update(self.time_sec)    # update ET
            if (self.DEBUG): print 'TF UPDATED Evap Component...'
        if (UPDATE_IP):
            self.ip.update(self.time_sec)    # update infil
            if (self.DEBUG): print 'TF UPDATED Infil Component...'
        if (UPDATE_IIP):
            self.iip.update(self.time_sec)   # update ice_vars
            if (self.DEBUG): print 'TF UPDATED Ice Component...'
            ## move update q0 into here from merged2.py
            ## check_infiltration()
        if (UPDATE_GP):
            self.gp.update(self.time_sec)   # update GW ?
            if (self.DEBUG): print 'TF UPDATED GW Component...'

        if (self.DEBUG): print 'MADE IT PAST updates ...'
        
        #---------------------------------------------
        # Note that u and d from previous time step
        # must be used on RHS of the equations here.
        #---------------------------------------------
        # self.cp.update(self.time_sec)  # (uses others, so after them)
        # if (DEBUG): print 'UPDATED Channels Component...'
        
##        OK = (self.cp.get_status() != 'failed')
##        if not(OK):
##            self.Q_outlet = float64(0)   # (why is this here?)
##            self.DONE = True
##            return

        #---------------------------------------------
        # (2/1/10) dp.update() is now called by the
        # channels component in update_flow_volume()
        #---------------------------------------------
        # dp.update() uses a setter to modify cp.vol
        #---------------------------------------------
        ## if (UPDATE_DP):
##        self.dp.update(self.time_sec)
##        if (self.DEBUG): print 'TF UPDATED Diversions Component...'
        
        #-------------------------------------------------
        # These tests use Q_outlet, Q_peak, Q_last, so
        # make sure these are defined outside first.
        #-------------------------------------------------
        # NB!  Need float64() here in order for the
        #      .copy() method (at self.Q_last =) to work.
        #--------------------------------------------------
        self.Q_outlet = float64(self.cp.get_scalar_double('Q_outlet'))
        ## self.Q_outlet = self.cp.get_scalar_double('Q_outlet')
        self.Q_peak   = self.cp.get_scalar_double('Q_peak')
        FINISHED      = self.check_finished()
        STEADY        = self.check_steady_state()
        FAILED        = (self.cp.get_status == 'failed')
        if (FAILED):
            if (self.DEBUG): print 'CHANNELS.update() failed.'
            self.status   = 'failed'
            self.Q_outlet = float64(0)   # (why is this here?)
        self.Q_last   = self.Q_outlet.copy()        
        self.DONE     = (FINISHED or STEADY or FAILED)
       
        #-------------------------
        # Increment the timestep
        #-------------------------
        self.update_time()
        self.status = 'updated'
        
    #   update()          
    #-------------------------------------------------------------            
    def finalize(self):

        self.status = 'finalizing'

        #--------------------------------------------
        # This is called by the update() method.
        # We'd like to force another output, but it
        # is now based on elapsed time.  10/29/11.
        #--------------------------------------------        
        # if (self.mode == 'driver'):
        #     self.print_time_and_value(self.Q_outlet, 'Q_out', '[m^3/s]')
 
        #-----------------------------------
        # Close input & output files, etc.
        #-----------------------------------
        # print 'CALLING cp.finalize()...'
        self.cp.finalize()
        # print 'CALLING sp.finalize()...'
        self.sp.finalize()
        # print 'CALLING mp.finalize()...'
        self.mp.finalize()
        # print 'CALLING ep.finalize()...'
        self.ep.finalize()
        # print 'CALLING ip.finalize()...'
        self.ip.finalize()
        # print 'CALLING gp.finalize()...'
        self.gp.finalize()
        # print 'CALLING dp.finalize()...'
        self.dp.finalize()
        # print 'CALLING iip.finalize()...'
        self.iip.finalize()

        print '======================='
        print 'Simulation complete.'
        print '======================='
        print ' '
        
        #----------------
        # Print reports
        #----------------      
##        self.print_final_report()  # (Now called in IMPL_file. 10/27/11)
##        self.print_mins_and_maxes(FINAL=True)   # (not ready yet)
##        self.print_uniform_precip_data()  # (not ready yet)
##        self.print_mass_balance_report()  # (not ready yet)

        #--------------------
        # Close the logfile
        #--------------------
        if (self.WRITE_LOG):
            ## print '###  Closing log file.'
            self.log_unit.close()

        #----------------------
        # Print final message
        #----------------------
        TF_Print('Finished.' + '  (' + self.case_prefix + ')')
        TF_Print(' ')
        self.status = 'finalized'

        #---------------------------
        # Release all of the ports
        #----------------------------------------
        # Make this call in "finalize()" method
        # of the component's CCA Imple file
        #----------------------------------------
        # self.release_cca_ports( port_names, d_services )
        
    #   finalize()
    #-------------------------------------------------------------------
    def get_cca_port_info(self):

        self.cca_port_names = ['meteorology', 'channels',
                               'snow', 'evap', 'infil', 'satzone',
                               'diversions', 'ice']
        self.cca_port_short = ['mp', 'cp', 'sp', 'ep', 'ip',
                               'gp', 'iip', 'dp']
        self.cca_port_type  = "IRFPort"
        self.cca_project    = "edu.csdms.models"

    #   get_cca_port_info()
    #-------------------------------------------------------------------
    def embed_child_components(self):

        #------------------------------------------------
        # Note: Don't call this if (self.CCA == True).
        #------------------------------------------------
        # Instantiate and embed "process components"
        # in the place of the CCA ports.
        #------------------------------------------------
        # But how do we choose a given process method
        # such as "kinematic wave" ??
        #------------------------------------------------
        import met_base
        import channels_kinematic_wave
        import snow_degree_day
        import ET_read_file
        # import ET_priestley_taylor
        import infil_green_ampt
        import GW_darcy_layers
        import diversions_fraction_method
        import ice_base

        self.mp  = met_base.met_component()
        self.cp  = channels_kinematic_wave.channels_component()
        self.sp  = snow_degree_day.snow_component()
        self.ep  = ET_read_file.ET_component()
        # self.ep  = ET_priestley_taylor.ET_component()
        self.ip  = infil_green_ampt.infil_component()
        self.gp  = GW_darcy_layers.GW_component()
        self.dp  = diversions_fraction_method.diversions_component()
        self.iip = ice_base.ice_component()   #########
        
    #   embed_child_components()    
    #-------------------------------------------------------------------
    def add_child_ports(self):

        self.add_child_port('cp', 'mp')
        self.add_child_port('cp', 'sp')
        self.add_child_port('cp', 'ep')
        self.add_child_port('cp', 'ip')
        self.add_child_port('cp', 'gp')
        self.add_child_port('cp', 'iip')
        self.add_child_port('cp', 'dp')
        #----------------------------------
        self.add_child_port('sp', 'mp')
        #----------------------------------
        self.add_child_port('ep', 'mp')
        self.add_child_port('ep', 'cp')
        self.add_child_port('ep', 'sp')
        self.add_child_port('ep', 'ip')
        self.add_child_port('ep', 'gp')
        #----------------------------------
        self.add_child_port('ip', 'mp')
        self.add_child_port('ip', 'sp')
        self.add_child_port('ip', 'ep')
        self.add_child_port('ip', 'gp')
        self.add_child_port('ip', 'cp')         ##### (7/7/10)
        #----------------------------------
        self.add_child_port('gp', 'ip')
        self.add_child_port('gp', 'cp')
        #----------------------------------
        self.add_child_port('dp', 'cp')
        #----------------------------------
        self.add_child_port('mp', 'sp')
        ## self.add_child_port('mp', 'pp')   ##### in future ?
        #----------------------------------
        self.add_child_port('iip', 'mp')
        self.add_child_port('iip', 'sp')
        
    #   add_child_ports()
    #-------------------------------------------------------------------
    def initialize_ports(self):
        
        #--------------------------------------------------
        # Initialize the process objects/components
        # This is also where output files are opened.
        #--------------------------------------------------
        DEBUG = True      
        
        if (self.mp.get_status() != 'initialized'):        # met vars
            if (DEBUG): print 'Initializing "meteorology"...'
            self.mp.initialize( cfg_prefix=self.cfg_prefix ) 
            if (DEBUG): print 'TopoFlow initialized METEOROLOGY.\n'

        # Need this after Channels component now (2/16/10)
##        if (self.dp.get_status() != 'initialized'):        # diversion vars
##            if (DEBUG): print 'Initializing "diversions"...'
##            self.dp.initialize( cfg_prefix=self.cfg_prefix ) 
##            if (DEBUG): print 'TopoFlow initialized DIVERSIONS.\n'
            
##        if (self.pp.get_status() != 'initialized'):        # precip vars
##            if (DEBUG): print 'Initializing "precipitation"...'
##            self.pp.initialize( cfg_prefix=self.cfg_prefix ) 
##            if (DEBUG): print 'TopoFlow initialized PRECIP.\n'
        
        if (self.cp.get_status() != 'initialized'):        # channel vars
            if (DEBUG): print 'Initializing "channels"...'
            self.cp.initialize( cfg_prefix=self.cfg_prefix ) 
            if (DEBUG): print 'TopoFlow initialized CHANNELS.\n'

        #------------------------------------------------------
        # (2/16/10) This needs cp.dt, and currently sets some
        # values in the Channels component when initialized.
        #------------------------------------------------------
        if (self.dp.get_status() != 'initialized'):        # diversion vars
            if (DEBUG): print 'Initializing "diversions"...'
            self.dp.initialize( cfg_prefix=self.cfg_prefix ) 
            if (DEBUG): print 'TopoFlow initialized DIVERSIONS.\n'
            
        if (self.sp.get_status() != 'initialized'):        # snow vars
            if (DEBUG): print 'Initializing "snow"...'
            self.sp.initialize( cfg_prefix=self.cfg_prefix ) 
            if (DEBUG): print 'TopoFlow initialized SNOW.\n'
        
        # Must be before ET init.
        if (self.gp.get_status() != 'initialized'):        # GW vars
            if (DEBUG): print 'Initializing "satzone"...'
            self.gp.initialize( cfg_prefix=self.cfg_prefix ) 
            if (DEBUG): print 'TopoFlow initialized SATZONE.\n'
        
        if (self.ep.get_status() != 'initialized'):        # ET vars
            if (DEBUG): print 'Initializing "evap"...'
            self.ep.initialize( cfg_prefix=self.cfg_prefix ) 
            if (DEBUG): print 'TopoFlow initialized EVAP.\n'
        
        if (self.ip.get_status() != 'initialized'):        # infil vars
            if (DEBUG): print 'Initializing "infil"...'
            self.ip.initialize( cfg_prefix=self.cfg_prefix ) 
            if (DEBUG): print 'TopoFlow initialized INFIL.\n'

        if (self.iip.get_status() != 'initialized'):        # ice vars
            if (DEBUG): print 'Initializing "ice"...'
            self.iip.initialize( cfg_prefix=self.cfg_prefix ) 
            if (DEBUG):
                print 'TopoFlow initialized ICE.\n'
                
        #--------------------------------------
        # Check if "self" has "iip" attribute
        #--------------------------------------
        # It must if we made it to here.
        #--------------------------------------
        ICE_PORT_TEST = False
        if (ICE_PORT_TEST):
            try:
                dt = self.get_port_data('dt', self.iip, 'ICE')
                print 'dt =', dt
            except:
                print 'ERROR: Could not get "dt" from ICE port.'
                print ' '
                ## return  # (try next block, too)

            try:
                MR = self.get_port_data('MR', self.iip)
                MR = numpy.float64(MR)
                print '(MRmin, MRmax) =', MR.min(), MR.max()
            except:
                print 'ERROR: Could not get "MR" from ICE port.'
                print ' '
                ## return
        
    #   initialize_ports()
    #-------------------------------------------------------------------
    def compute_update_steps(self):

        #---------------------------------------------------
        # NB! TopoFlow uses channel process dt as main dt.
        #---------------------------------------------------
        # Compute an "update step" for each process [secs]
        #---------------------------------------------------
        ## pp_dt  = self.pp.get_scalar_double('dt')
        mp_dt  = self.mp.get_scalar_double('dt')
        sp_dt  = self.sp.get_scalar_double('dt')
        ep_dt  = self.ep.get_scalar_double('dt')
        ip_dt  = self.ip.get_scalar_double('dt')
        gp_dt  = self.gp.get_scalar_double('dt')
        dp_dt  = self.dp.get_scalar_double('dt')
        iip_dt = self.iip.get_scalar_double('dt')  ## VARIABLE DT POSSIBLE
        
        ## self.pp_update_step = ceil(pp_dt / self.dt).astype('Int32')        
        self.mp_update_step = ceil(mp_dt / self.dt).astype('Int32')
        self.sp_update_step = ceil(sp_dt / self.dt).astype('Int32')
        self.ep_update_step = ceil(ep_dt / self.dt).astype('Int32')
        self.ip_update_step = ceil(ip_dt / self.dt).astype('Int32')
        self.gp_update_step = ceil(gp_dt / self.dt).astype('Int32')
        self.dp_update_step = ceil(dp_dt / self.dt).astype('Int32')
        self.iip_update_step= ceil(iip_dt /self.dt).astype('Int32')
        
    #   compute_update_steps()
    #-------------------------------------------------------------            
    def check_finished(self):

        #------------------------------------------------------- 
        # Notes: Don't set "Q_last" at bottom of this function
        #        because it is also used by other stop tests.
        #------------------------------------------------------- 
        
        #---------------------------------
        # Methods to stop the simulation
        #---------------------------------
        if (self.stop_method == 'Q_peak_fraction'):
            #----------------------------------------------------
            # Run until the outlet hydrograph drops to less
            # than "Qp_fraction" of the peak value before that.
            #----------------------------------------------------  
            FALLING_LIMB = (self.Q_last > self.Q_outlet)
            
##            if (FALLING_LIMB):  print "ON FALLING LIMB..."
##            print 'self.Q_last  =', self.Q_last
##            print 'self.Q_outlet =', self.Q_outlet
##            print ' '
            
            #--------------------------------------------------------
            # With DYNAMIC_WAVE, it is possible for some reason for
            # Q_outlet to drop back to zero early in the simulation
            # so that model run ends right away.  (2/13/07)
            # Uncomment the debugging section below.
            #--------------------------------------------------------
            if (FALLING_LIMB):
                Q_stop   = (self.Q_peak * self.Qp_fraction)
                self.DONE = (self.Q_outlet <= Q_stop) and \
                            (self.Q_outlet > 0)

            if (self.DONE):
                print 'Stopping: Reached Q_peak fraction.\n'
                
##            print 'FALLING_LIMB   =', FALLING_LIMB
##            print 'Q_last         =', self.Q_last
##            print 'Q_peak         =', self.Q_peak
##            print 'Qpeak_fraction =', self.Qp_fraction
##            print ' '
            
                #--------------
                # For testing
                #--------------
                #if (DONE):
                #    print 'Q_last         =', self.Q_last
                #    print 'Q_peak         =', self.Q_peak
                #    print 'Qpeak_fraction =', self.Qp_fraction
                #    print 'Q[self.outlet_ID]   =', Q_outlet
                #    print 'Q_stop         =', Q_stop
                #    print ' '

        elif (self.stop_method == 'Until_model_time'):
            #--------------------------------------------------
            # Run until specified "stopping time", in minutes
            #--------------------------------------------------
            self.DONE = (self.time_min >= self.T_stop)  # [minutes]
            if (self.DONE):
                print 'Stopping: Reached stop time', self.T_stop, '\n'
        elif (self.stop_method == 'Until_n_steps'):
            #--------------------------------------
            # Run for a specified number of steps
            #--------------------------------------
            self.DONE = (self.time_index >= self.n_steps)
            if (self.DONE):
                print 'Stopping: Reached number of steps.\n'
        else:
            raise RuntimeError('No match found for expression')

        return self.DONE
      
    #   check_finished()
    #-------------------------------------------------------------------
    def check_steady_state(self):

        #-------------------------------------------------------
        # Notes:  See "initialize_stop_vars()" for definitions
        #         of steady_tol, nonzero_tol, n_same_max and
        #         Q_last.
        #-------------------------------------------------------
        STEADY = False

        #------------------------------------------
        # Count number of steps with same Q-value
        #------------------------------------------
        delta_Q = absolute(self.Q_outlet - self.Q_last)
        if ( delta_Q <= self.steady_tol ):
            ## print '(time_index, dQ) =', self.time_index, delta_Q
            self.n_same += 1
        else:
            self.n_same  = 0
        
        #------------------------------------
        # Check for steady-state, with Q > 0
        #------------------------------------
        if (self.stop_method == 'Q_peak_fraction') and \
           (self.Q_outlet > self.nonzero_tol) and \
           (self.n_same   > self.n_same_max):

            STEADY = True
            msg = ['-----------------------------------------------------------', \
                   'WARNING: It appears that discharge, Q, has reached', \
                   '         a steady-state condition.', \
                   '         Discharge at outlet near: ' + str(self.Q_outlet), \
                   '             for ' + str(self.n_same) + ' timesteps.', \
                   '         Aborting model run.', \
                   '-----------------------------------------------------------', \
                   ' ']
                   ### 'Do you want to continue anyway ?', ' '])
            for line in msg:
                print line

            
            ## answer = GUI_Message(msg, QUESTION=True)
            ## DONE = (answer.lower() == 'no')
            ## if not(DONE):    
            ##     n_same = int32(0)
            ##
            #TF_Print,'****************************************************'
            #TF_Print,'Aborting model run: '
            #TF_Print,'Discharge, Q, has reached steady-state.'
            #TF_Print,'****************************************************'
            #msg = [ $
            #'WARNING:  Route_Flow aborted.', ' ',$
            #'Discharge, Q, has reached steady-state. ', ' ']
            #GUI_Error_Message, msg
            #STEADY = True
        
        #-----------------------------------------------
        # (3/20/07) Commented out, since user can now
        # press any key to stop the model run.  Note
        # that Q-value can remain zero for a long time
        # if everything is infiltrating or snow depth
        # is building, etc.
        #-----------------------------------------------
        # Check for unchanging Q-value of zero ?
        #-----------------------------------------
        #if (STOP_METHOD eq 0b) AND (Q_peak eq 0.0) AND (n_same gt nn) then begin
        #    msg = [' ', $
        #    'ERROR: ', ' ', $
        #    'Discharge at outlet is zero for all times. ', ' ',$
        #    'Is there a runoff-producing process ? ', ' ']
        #    TF_Print,'*************************************************'
        #    TF_Print, msg[1]
        #    TF_Print, msg[3]
        #    TF_Print,'*************************************************'
        #    TF_Print,' '
        #    GUI_Error_Message, msg
        #    DONE = 1b
        #endif
          
        return STEADY
    
    #   check_steady_state
    #-------------------------------------------------------------
    def check_interrupt(self):

        #------------------------------------------------------------
        #( 3/21/07) This method checks for a keyboard interrupt
        # after a fixed amount of real time has elapsed.  This
        # works much better than using something like (n mod n_check)
        # since it avoids checking fast runs too often (which slows
        # down performance) or slow runs too seldom (which means the
        # user can't interrupt the run).  It only checks the elapsed
        # time once each time through the loop, however, so the time
        # required for one pass imposes a lower bound.
        #------------------------------------------------------------
        elapsed_time = (time.time() - self.last_check_time)
        if (elapsed_time > 2):    
            #print,'****************************************************'
            #print,'Checking interrupt: n = ' + TF_String(n)
            #print,'****************************************************'

            ########################################
            ##  REPLACE WITH THE CODE ITSELF
            ########################################
            ## Check_Interrupt(STOP_ID, OK)
            OK = True  # (over-ridden, for now)
            
            if not(OK):
                self.finalize()
                TF_Print(' ')
                TF_Print('----------------------------------')
                TF_Print(' Simulation terminated by user.')
                TF_Print('----------------------------------')
                return

            self.last_check_time = time.time()

    #   check_interrupt()
    #-------------------------------------------------------------
    def initialize_time_vars(self):

        #------------------
        # Start the clock
        #------------------
        self.start_time = time.time()

        #--------------------------------
        # Initialize the time variables
        #--------------------------------        
        self.time_units = 'seconds'
        self.time_index = int32(0)
        self.time       = float64(0)
        self.DONE       = False
        
        self.time_sec   = float64(0)
        self.time_min   = float64(0)

        self.last_check_time  = time.time()  # (for check_interrupt() )
        self.last_print_time  = time.time()  # (for print_time_and_value() )
        self.last_plot_time   = float64(0)   # (for update_hydrograph_plot() )

        #---------------------------------------
        # Set the model run timestep to that
        # of the "channel_flow" process [secs]
        #---------------------------------------
        self.dt = self.cp.get_scalar_double('dt')
        
    #   initialize_time_vars()
    #-------------------------------------------------------------
    def initialize_stop_vars(self):

        TF_Print('Setting stop method to: ' + self.stop_method)

        #---------------------------------------------------------
        # Define some constants for "check_steady_state() method
        #----------------------------------------------------------------
        # Best value of tolerance also depends on the time step.
        # For "plane" case, result changed with timestep = 2 or 4 secs.
        #----------------------------------------------------------------
        # "Optimal" value of steady_tol was found by trial and error.
        #----------------------------------------------------------------        
        self.steady_tol  = float64(1E-5)
        self.nonzero_tol = float64(1E-5)
        #self.nonzero_tol = 1e-6      #(worked better for "plane" case with step=2s)
        self.n_same      = int32(0)
        self.n_same_max  = int32(499) # (a number of time steps)
        self.Q_last      = float64(0)
        
        if (self.stop_method == 'Q_peak_fraction'):
            #----------------------------------------------------
            # Run until the outlet hydrograph drops to less
            # than "Qp_fraction" of the peak value before that.
            #----------------------------------------------------
            T_stop = 0
            Tstr   = '  [min]'
        elif (self.stop_method == 'Until_model_time'):
            #--------------------------------------------------
            # Run until specified "stopping time", in minutes
            #--------------------------------------------------
            T_stop = self.T_stop_model
            mstr   = ('%10.2f' % T_stop)
            Tstr   = ' of ' + mstr + '  [min]'
        elif (self.stop_method == 'Until_n_steps'):
            #--------------------------------------
            # Run for a specified number of steps
            #--------------------------------------
            n_steps = self.n_steps
            T_stop  = (n_steps * self.dt / float64(60))   #[sec] -> [min]
            mstr    = ('%10.2f' % T_stop)
            Tstr   = ' of ' + mstr + '  [min]'
        else:
            print 'ERROR: Invalid stopping method.'
            return

        self.T_stop = T_stop
        self.Tstr   = Tstr
            
    #   initialize_stop_vars()       
    #-------------------------------------------------------------
##    def initialize_gui(self):
##        
##        #-----------------------------------------
##        # Set various widget IDs and info
##        # (should this be called by __init__ ??)
##        #-----------------------------------------
##        self.leader_ID  = int32(0)
##        self.base_ID    = int32(0)
##        self.start_ID   = int32(0)
##        self.stop_ID    = int32(0)
##        self.draw_ID    = int32(0)
##        self.win_num    = int32(0)   ########
##        self.temp_winID = int32(-1)
##        self.npanels    = npanels
##        self.panel_IDs  = zeros([npanels], dtype='Int32')
##        
##        self.stop_method_ID = int32(0)  #####
##
##        #----------------------------------------
##        # Option to plot hydrograph dynamically
##        #----------------------------------------
##        if (self.PLOT):
##            self.initialize_hydrograph_plot()
##
##    #   initialize_gui()
    #-------------------------------------------------------------
##    def initialize_hydrograph_plot(self):
##        
##        #------------------------------------------------------
##        # NB! Get the window number for the draw widget.
##        #     This assumes that a small draw window has been
##        #     added in the output log panel.  See the function
##        #     called GUI_Stopping in GUI_main.pro.
##        #------------------------------------------------------
##        # NB! Rainbow + white color table was loaded earlier
##        #     by GUI_Stopping, so black=0 and white=255.
##        #------------------------------------------------------
##        ## Initialize_Hydrograph_Window(DRAW_ID, win_num)
##        
##        self.np = int32(0)
##        self.np_init = int16(1000)
##        np_max = self.np_init
##        self.tvals = zeros([np_max], dtype='Float32')
##        self.Qvals = zeros([np_max], dtype='Float32')  
##
##    #   initialize_hydrograph_plot()
    #-------------------------------------------------------------
##    def initialize_mass_totals(self):
## 
##        #-------------------------------------------------------
##        # Prepare to track total mass of each process for the
##        # entire DEM.  This will actually be a volume, in m^3.
##        #-------------------------------------------------------
##        # Volume into and out of main basin is tracked by
##        # data members of the "basins" class
##        #-------------------------------------------------------        
##        self.vol_P  = float64(0)
##        self.vol_SM = float64(0)
##        self.vol_IN = float64(0)
##        self.vol_ET = float64(0)
##        self.vol_GW = float64(0)
##        self.vol_R  = float64(0)
##        self.vol_Rg = float64(0)
##
##    #   initialize_mass_totals()       
##    #-------------------------------------------------------------
##    def update_mass_totals(self):
##
##        N_P = N_S = N_I = N_E = N_G = N_R = N_Rg = int32(1)
##        
##        if (size(self.pp.P)  == 1): N_P  = self.n_pixels
##        if (size(self.sp.SM) == 1): N_S  = self.n_pixels
##        if (size(self.ip.IN) == 1): N_I  = self.n_pixels
##        if (size(self.ip.Rg) == 1): N_Rg = self.n_pixels
##        if (size(self.ep.ET) == 1): N_E  = self.n_pixels
##        if (size(self.gp.GW) == 1): N_G  = self.n_pixels
##        if (size(self.R)     == 1): N_R  = self.n_pixels
##
##        mfac = self.dt * self.da
##        
##        if (size(self.da) == 1):    
##            self.vol_P  += sum(double(self.pp.P   * mfac)) * N_P
##            self.vol_SM += sum(double(self.sp.SM  * mfac)) * N_S
##            self.vol_IN += sum(double(self.ip.IN  * mfac)) * N_I
##            self.vol_Rg += sum(double(self.ip.Rg  * mfac)) * N_Rg            
##            self.vol_ET += sum(double(self.ep.ET  * mfac)) * N_E
##            self.vol_GW += sum(double(self.gp.GW  * mfac)) * N_G
##            self.vol_R  += sum(double(self.R      * mfac)) * N_R
##
##        else:    
##            #----------------------------------------
##            # Note:  da is a grid so mfac is a grid
##            #----------------------------------------
##            self.vol_P  += sum(double(self.pp.P   * mfac))
##            self.vol_SM += sum(double(self.sp.SM  * mfac))
##            self.vol_IN += sum(double(self.ip.IN  * mfac))
##            self.vol_Rg += sum(double(self.ip.Rg  * mfac))
##            self.vol_ET += sum(double(self.ep.ET  * mfac))
##            self.vol_GW += sum(double(self.gp.GW  * mfac))
##            self.vol_R  += sum(double(self.R      * mfac))
##
##    #   update_mass_totals
    #-------------------------------------------------------------
    def update_hydrograph_plot(self):
        
        #-----------------------------------------
        # Plot hydrograph up until now (3/22/07)
        #-----------------------------------------
        # plus sign (psym=1), point (psym=3)
        #-----------------------------------------
        elapsed_plot_time = (self.time_min - self.last_plot_time)
        if (self.PLOT and (elapsed_plot_time > 1.0)):    
            #------------------------------------
            # Report an "instantaneous" Q value
            #------------------------------------
            ########################
            np      = self.np
            np_init = self.np_init
            tvals   = self.tvals
            Qvals   = self.Qvals
            ########################
            Q_main_out = Pixel_Var(self.cp.Q, self.outlet_ID)
            self.tvals[np] = self.time_min
            self.Qvals[np] = Q_main_out
            matplotlib.pyplot.figure(self.win_num + 1)
            matplotlib.pyplot.plot(tvals[0:np+1], Qvals[0:np+1], \
                                   color='k', marker='.')
            matplotlib.pyplot.axes(axisbg='w')
            matplotlib.pyplot.ylim(array(0, float32(1.03) * Qvals.max()))
            matplotlib.pyplot.show()  #**** -1
            time.sleep(float32(0.005))  #[seconds]
            np = (np + int32(1))
            if (np == self.np_max):
                ## Use numpy.concatenate() here ??
                tvals = array([tvals, zeros([np_init], dtype='Float32')])
                Qvals = array([Qvals, zeros([np_init], dtype='Float32')])
                self.np_max = size(tvals)
            self.last_plot_time = self.time_min
            ########################
            self.np     = np
            self.tvals  = tvals
            self.Qvals  = Qvals
            ########################

    #   update_hydrograph_plot()
    #-------------------------------------------------------------
    def print_final_report(self, comp_name='TopoFlow',
                           mode='nondriver'):

        #------------------------------------------------------
        # NB! This overrides CSDMS_base.print_final_report(),
        # so it must have the same arguments. (10/27/11)
        #------------------------------------------------------
        
        #------------------------------------
        # Gather information for the report
        #------------------------------------
        NAME      = self.site_prefix
        COMMENT   = self.comment
        T_stop    = self.T_stop
        Q_final   = self.Q_outlet 
        T_final   = self.time_min
        outlet_ID = self.outlet_ID
        n_steps   = self.time_index
        # outlet_col = (outlet_ID % self.nx)   ## (may need this)
        # outlet_row = (outlet_ID / self.nx)
        
        volume_out = self.cp.get_scalar_double('volume_out')
        Q_peak     = self.cp.get_scalar_double('Q_peak')
        T_peak     = self.cp.get_scalar_double('T_peak')
        u_peak     = self.cp.get_scalar_double('u_peak')
        Tu_peak    = self.cp.get_scalar_double('Tu_peak')
        d_peak     = self.cp.get_scalar_double('d_peak')
        Td_peak    = self.cp.get_scalar_double('Td_peak')
        sample_dt  = self.cp.get_scalar_double('save_pixels_dt')   ## (needed ?)
        
        ## volume_in  = self.bp.volume_in
        ## basin_area = self.bp.get_scalar_double('basin_area')
        basin_area = self.bp.basin_area
        TRACK_VOLUME = False   ################ (not ready yet) ###############
        volume_in    = float64(0)

        P_max = self.mp.get_scalar_double('P_max')
       
        mp_dt = self.mp.get_scalar_double('dt') 
        cp_dt = self.cp.get_scalar_double('dt')
        sp_dt = self.sp.get_scalar_double('dt')
        ep_dt = self.ep.get_scalar_double('dt')
        ip_dt = self.ip.get_scalar_double('dt')
        gp_dt = self.gp.get_scalar_double('dt')
        iip_dt= self.iip.get_scalar_double('dt')
 
        #----------------------------
        # Construct run time string
        #----------------------------
        run_time = (time.time() - self.start_time)
        if (run_time > 60):
            run_time = run_time / float64(60)
            rt_units = ' [min]'
        else:
            rt_units = ' [sec]'
        run_time_str = str(run_time) + rt_units
        
        if (TRACK_VOLUME):  
            if (volume_in != 0):    
                percent_out = float32(100.0) * volume_out / volume_in
            else:    
                percent_out = float32(0.0)
        
        #-------------------
        # Print the report
        #-------------------
        hline = ''.ljust(60, '-')
        TF_Print(hline)
        TF_Print(tf_utils.TF_Version())
        TF_Print(time.asctime())  #####
        TF_Print(' ')
        # TF_Print('Simulated Hydrograph for ' + NAME)
        TF_Print('Input directory:      ' + self.in_directory)
        TF_Print('Output directory:     ' + self.out_directory)
        TF_Print('Site prefix:          ' + self.site_prefix)
        TF_Print('Case prefix:          ' + self.case_prefix)
        if (COMMENT is not None):
            TF_Print(' ')
            TF_Print(COMMENT)
        
        TF_Print(' ')
        TF_Print('Simulated time:      ' + str(T_final) + ' [min]')
        TF_Print('Program run time:    ' + run_time_str)
        TF_Print(' ')
        TF_Print('Number of timesteps: ' + str(n_steps))
        TF_Print('Number of columns:   ' + str(self.nx) )
        TF_Print('Number of rows:      ' + str(self.ny) )
        TF_Print(' ')
        TF_Print('Channel timestep:    ' + str(cp_dt) + ' [s]')
        TF_Print('Precip/met timestep: ' + str(mp_dt) + ' [s]')
        TF_Print('Snowmelt timestep:   ' + str(sp_dt) + ' [s]')
        TF_Print('ET timestep:         ' + str(ep_dt) + ' [s]')
        TF_Print('Infiltr. timestep:   ' + str(ip_dt) + ' [s]')
        TF_Print('Sat. zone timestep:  ' + str(gp_dt) + ' [s]')
        TF_Print('Icemelt timestep:    ' + str(iip_dt) + ' [s]')
        
        #** TF_Print,'Overland timestep: ' + str(op_dt) + ' [s]'
        #** TF_Print,'Sampled timestep:  ' + str(sample_dt) + ' [s]'

        if (self.stop_method == 'Until_model_time'):    
            TF_Print('T_stop:            ' + str(T_stop) + ' [min]')
        TF_Print(' ')
        TF_Print('Main outlet ID:    ' + str(outlet_ID) + ' (row, col)')
        # TF_Print('Outlet (col,row):  ' + TF_String(out_col) + ', ' + TF_String(out_row))
        TF_Print('Basin_area:        ' + str(basin_area) + ' [km^2] ')
        #*** TF_Print,'Basin_length:      ' + TF_String(basin_length) + ' [m]'
        TF_Print(' ')

        MANNING     = self.cp.get_scalar_long('MANNING')
        LAW_OF_WALL = self.cp.get_scalar_long('LAW_OF_WALL')
        
        if (MANNING):     
            nmin = self.cp.get_scalar_double('nval_min')   ####
            nmax = self.cp.get_scalar_double('nval_max')   ####
            TF_Print("Min Manning's n:   " + str(nmin))
            TF_Print("Max Manning's n:   " + str(nmax))
        
        if (LAW_OF_WALL):    
            z0min = self.cp.get_scalar_double('z0val_min')  ####
            z0max = self.cp.get_scalar_double('z0val_max')  ####
            TF_Print("Min z0 value:      " + str(z0min) + ' [m]')
            TF_Print("Max z0 value:      " + str(z0max) + ' [m]')
     
        TF_Print(' ')
        TF_Print('Q_final:           ' + str(Q_final)         + ' [m^3/s]')
        TF_Print('Q_peak:            ' + str(Q_peak)  + ' [m^3/s]')
        TF_Print('T_peak:            ' + str(T_peak)  + ' [min]')
        TF_Print('u_peak:            ' + str(u_peak)  + ' [m/s]')
        TF_Print('Tu_peak:           ' + str(Tu_peak) + ' [min]')
        TF_Print('d_peak:            ' + str(d_peak)  + ' [m]')
        TF_Print('Td_peak:           ' + str(Td_peak) + ' [min]')
        TF_Print(' ')
        
        if (TRACK_VOLUME):    
            TF_Print('Total volume out:  ' + str(volume_out) + ' [m^3]')
            if (basin_area != 0):    
                TF_Print('Rain volume in:    ' + str(volume_in)   + ' [m^3]')
                TF_Print('Percentage out:    ' + str(percent_out) + ' [%]')
                TF_Print(' ')
                
        #--------------------------------
        # Print the maximum precip rate
        #--------------------------------
        MPR = (P_max * self.mps_to_mmph)   # ([m/s] -> [mm/hr])
        TF_Print('Max(precip rate):  ' + str(MPR) + ' [mm/hr]')
        TF_Print(' ')

        #---------------------
        # Write to logfile ?
        #---------------------
        if not(self.WRITE_LOG): return
        log_unit = self.log_unit
        log_unit.write("\n")
        log_unit.write(tf_utils.TF_Version() + "\n")
        log_unit.write(time.asctime() + "\n")  #####
        log_unit.write("\n")
        # log_unit.write('Simulated Hydrograph for ' + NAME + "\n")
        log_unit.write('Input directory:      ' + self.in_directory  + "\n")
        log_unit.write('Output directory:     ' + self.out_directory + "\n")
        log_unit.write('Site prefix:          ' + self.site_prefix   + "\n")
        log_unit.write('Case prefix:          ' + self.case_prefix   + "\n")
        if (COMMENT is not None):
            log_unit.write("\n")
            log_unit.write(COMMENT)
        log_unit.write("\n")
        log_unit.write('Simulated time:      ' + str(T_final)    + " [min]\n")
        log_unit.write('Program run time:    ' + run_time_str + "\n")
        log_unit.write("\n")
        log_unit.write('Number of timesteps: ' + str(n_steps) + "\n")
        log_unit.write('Number of columns:   ' + str(self.nx) + "\n" )
        log_unit.write('Number of rows:      ' + str(self.ny) + "\n" )
        log_unit.write('\n')
        log_unit.write('Channel timestep:    ' + str(cp_dt) + " [s]\n")
        log_unit.write('Precip/met timestep: ' + str(mp_dt) + " [s]\n")    
        log_unit.write('Snowmelt timestep:   ' + str(sp_dt) + " [s]\n")
        log_unit.write('ET timestep:         ' + str(ep_dt) + " [s]\n")
        log_unit.write('Infiltr. timestep:   ' + str(ip_dt) + " [s]\n")
        log_unit.write('Sat. zone timestep:  ' + str(gp_dt) + " [s]\n")
        log_unit.write('Icemelt timestep:    ' + str(iip_dt) + " [s]\n")
        #** printf,LU,'Overland timestep:  ' + str(op_dt) + " [s]\n"
        #** printf,LU,'Sampled timestep:   ' + str(sample_dt)  + " [s]\n"
        
        if (self.stop_method == 'Until_model_time'):    
            log_unit.write('T_stop:            ' + str(T_stop) + " [min]\n")
        log_unit.write("\n")
        log_unit.write('Main Outlet ID:    ' + str(outlet_ID) + "(row, col)\n")
        # log_unit.write('Outlet (col,row):  ' + str(out_col) + ', ' + str(out_row) + "\n")
        log_unit.write('Basin_area:        ' + str(basin_area)   + " [km^2]\n")
        #*** log_unit.write('Basin_length:      ' + str(basin_length) + " [m]\n")
        log_unit.write("\n")
        
        if (MANNING):    
            # See above: nmin = min(nval, max=nmax)
            log_unit.write("Min Manning's n:   " + str(nmin) + "\n")
            log_unit.write("Max Manning's n:   " + str(nmax) + "\n")
        
        if (LAW_OF_WALL):    
            # See above: z0min = min(z0val, max=z0max)
            log_unit.write("Min z0 value:      " + str(z0min) + " [m]\n")
            log_unit.write("Max z0 value:      " + str(z0max) + " [m]\n")
        log_unit.write("\n")
        
        log_unit.write('Q_final:           ' + str(Q_final) + " [m^3/s]\n")
        log_unit.write('Q_peak:            ' + str(Q_peak)  + " [m^3/s]\n")
        log_unit.write('T_peak:            ' + str(T_peak)  + " [min]\n")
        log_unit.write('u_peak:            ' + str(u_peak)  + " [m/s]\n")
        log_unit.write('Tu_peak:           ' + str(Tu_peak) + " [min]\n")
        log_unit.write('d_peak:            ' + str(d_peak)  + " [m]\n")
        log_unit.write('Td_peak:           ' + str(Td_peak) + " [min]\n")
        log_unit.write("\n")
        
        if (TRACK_VOLUME):    
            log_unit.write('Total volume out:  ' + str(volume_out) + " [m^3]\n")
            if (basin_area != 0):    
                log_unit.write('Rain volume in:    ' + str(volume_in)   + " [m^3]\n") 
                log_unit.write('Percentage out:    ' + str(percent_out) + " [%]\n")
                log_unit.write("\n")

        log_unit.write('Max(precip rate):  ' + str(MPR) + " [mm/hr]\n")
        log_unit.write("\n")
        
    #   print_final_report()
    #-------------------------------------------------------------
    def print_mins_and_maxes(self, FINAL=False):

        nx = self.nx
        ny = self.ny
        
        #-----------------------------
        # Compute the mins and maxes
        #-----------------------------
        Qmin = self.cp.Q[1:(ny - 2)+1,1:(nx - 2)+1].min()
        Qmax = self.cp.Q[1:(ny - 2)+1,1:(nx - 2)+1].max()
        umin = self.cp.u[1:(ny - 2)+1,1:(nx - 2)+1].min()
        umax = self.cp.u[1:(ny - 2)+1,1:(nx - 2)+1].max()
        dmin = self.cp.d[1:(ny - 2)+1,1:(nx - 2)+1].min()
        dmax = self.cp.d[1:(ny - 2)+1,1:(nx - 2)+1].max()
        
        f1 = '(F14.4)'   #(2/12/07)
        
        Qstr = TF_String(Qmin, FORMAT=f1)
        Qstr = Qstr + ', ' + TF_String(Qmax, FORMAT=f1)
        
        ustr = TF_String(umin, FORMAT=f1)
        ustr = ustr + ', ' + TF_String(umax, FORMAT=f1)
        dstr = TF_String(dmin, FORMAT=f1)
        dstr = dstr + ', ' + TF_String(dmax, FORMAT=f1)
        
        if (FINAL):    
            TF_Print('Final grid mins and maxes:')
        else:    
            TF_Print('------------------------------------------')
        TF_Print('Min(Q), Max(Q):   ' + Qstr + ' [m^3/s]')
        TF_Print('Min(u), Max(u):   ' + ustr + ' [m/s]')
        TF_Print('Min(d), Max(d):   ' + dstr + ' [m]')
        TF_Print(' ')
        
        #----------------------
        # Write to log file ?
        #----------------------
        if (self.WRITE_LOG):
            log_unit = self.log_unit
            if (FINAL):    
                log_unit.write("Final grid mins and maxes:\n")
            else:    
                log_unit.write("------------------------------------------\n")
            log_unit.write('Min(Q), Max(Q):   ' + Qstr + " [m^3/s]\n")
            log_unit.write('Min(u), Max(u):   ' + ustr + " [m/s]\n")
            log_unit.write('Min(d), Max(d):   ' + dstr + " [m]\n")
            log_unit.write("\n")
        
    #   print_mins_and_maxes()
    #-------------------------------------------------------------
    def print_uniform_precip_data(self):

        ## precip_method = self.pp.method
        precip_method = 2
        
        #---------------------------------------------------
        # Precip method 1 is special and spatially uniform
        #---------------------------------------------------
        if (precip_method == 1):
            rates = (self.pp.method1_rates * self.mps_to_mmph)    #[m/s] -> [mm/hr]
            durs  = self.pp.method1_durations      
            nr = size(rates)

            if (nr > 1): 
                rstr = str(rates[0])
                dstr = str(durs[0])
                for m in xrange(1, nr):
                    rstr += ('  ' + str(rates[m]))
                    dstr += ('  ' + str(durs[m]))
            else:
                rstr = TF_String(rates)
                dstr = TF_String(durs)
        elif (precip_method == 2) and \
             (self.pp.rate_type == 0) and \
             (self.pp.duration_type == 0):
                rstr = TF_String(self.pp.rate * self.mps_to_mmph)
                dstr = TF_String(self.pp.duration)
        else:
            #------------------------------------------------
            # Could have uniform precip with method 2 where
            # values are stored in a text file and read in
            # one by one.  Could read that file here.
            #------------------------------------------------
            return
            
        #-------------------------------------
        # This is too verbose in most cases?
        #------------------------------------- 
        #TF_Print,'Uniform precip. rate information: '
        #TF_Print,'Precip. rate:     ' + rstr + ' [mm/hr]'
        #TF_Print,'Duration:         ' + dstr + ' [min]'
        #TF_Print,' '
        
        #----------------------
        # Write to log file ?
        #----------------------
        if (self.WRITE_LOG):
            log_unit = self.log_unit
            log_unit.write("Uniform precip. rate information: \n")
            log_unit.write('Precip. rate:     ' + rstr + " [mm/hr]\n")
            log_unit.write('Duration:         ' + dstr + " [min]\n")
            log_unit.write("\n")
        
    #   print_uniform_precip_data()
    #-------------------------------------------------------------
    def print_dimless_number_data(self, basin_length):
 
##        rates     = self.pp.rates
##        durations = self.pp.durations
        
        rates     = self.mp.P
        durations = self.mp.dt
        
        nd        = size(durations)
        T_peak    = self.T_peak
        log_unit  = self.log_unit
        
        #-------------------------------------
        # Compute some dimensionless numbers
        # that characterize the hydrograph
        #-------------------------------------
        TAU_P = (self.T_peak / durations)
        TAU_F = (self.T_final / durations)
        PSI   = (basin_length / (rates[0] * durations[0] * float64(60)))
        
        tpstr = str(TAU_P[0])
        tfstr = str(TAU_F[0])
        for m in xrange(1, nd):
            tpstr = (tpstr + '  ' + str(TAU_P[m]))
            tfstr = (tfstr + '  ' + str(TAU_F[m]))
        
        #----------------------------------------
        # Make predictions for Q_peak and T_peak
        #-----------------------------------------
        Q_peak_pred = (float64(0.2) * rates[0] * (basin_length) ** float64(2) / float64(3))
        T_peak_pred = (float64(3) * durations[0])
        
        TF_Print('Dimensionless number information:\n')
        TF_Print('T_peak /Duration: ' + tpstr + "\n")
        TF_Print('T_final/Duration: ' + tfstr + "\n") 
        TF_Print('Psi=L/(R*TD):     ' + str(PSI) + ' [unitless]\n')
        TF_Print(' ')
        TF_Print('Q_peak predicted: ' + str(Q_peak_pred) + ' [m^3/s]\n')
        TF_Print('T_peak predicted: ' + str(T_peak_pred) + ' (min]\n')
        TF_Print(' ')
        
        #----------------------
        # Write to log file ?
        #----------------------
        if (self.WRITE_LOG):    
            self.log_unit.write('Dimensionless number information:\n')   
            self.log_unit.write('T_peak /Duration: ' + tpstr + "\n")
            self.log_unit.write('T_final/Duration: ' + tfstr + "\n")
            self.log_unit.write("\n")
            #printf,LU,'Psi=L/(R*TD):     ' + strPSI) + ' [unitless]'
            #printf,LU,' '
            #printf,LU,'Q_peak predicted: ' + str(Q_peak_pred) + ' [m^3/s]'
            #printf,LU,'T_peak predicted: ' + str(T_peak_pred) + ' [min]'
            #printf,LU,' '
                  
    #   print_dimless_number_data()
    #-------------------------------------------------------------
    def print_mass_balance_report(self):            

        ## vol_P  = self.pp.get_scalar_double('vol_P')
        
        vol_P  = self.mp.get_scalar_double('vol_P')
        vol_SM = self.sp.get_scalar_double('vol_SM')
        vol_IN = self.ip.get_scalar_double('vol_IN')
        vol_Rg = self.ip.get_scalar_double('vol_Rg')
        vol_ET = self.ep.get_scalar_double('vol_ET')
        vol_GW = self.gp.get_scalar_double('vol_GW')
        vol_R  = self.cp.get_scalar_double('vol_R')
        
        TF_Print('Volume rain        = ' + str(vol_P)  + ' [m^3]')
        TF_Print('Volume snowmelt    = ' + str(vol_SM) + ' [m^3]')
        TF_Print('Volume infiltrated = ' + str(vol_IN) + ' [m^3]')
        TF_Print('Volume evaporated  = ' + str(vol_ET) + ' [m^3]')
        TF_Print('Volume seepage     = ' + str(vol_GW) + ' [m^3]')
        TF_Print('Volume runoff      = ' + str(vol_R)  + ' [m^3]')
        TF_Print('Volume bottom loss = ' + str(vol_Rg) + ' [m^3]')
        TF_Print(' ')

        RICHARDS = self.ip.get_scalar_long('RICHARDS')
        print 'In topoflow.print_mass_balance_report(),'
        print '   RICHARDS =', RICHARDS
        print ' '
        if (RICHARDS):    
            #----------------------------------------
            # Richards' equation for infiltration
            # Ground water losses not counted above
            #----------------------------------------
            da         = self.da
            n_pixels   = self.n_pixels   #####
            vol_stored = float64(0)
            SCALAR_dz  = (size(self.ip.dz) == 1)
            SCALAR_da  = (size(da) == 1)     #(assume da is scalar or grid)
            #-----------------------------------
            dim_q  = numpy.ndim(self.ip.q)
            dim_qi = numpy.ndim(self.ip.qi)
            #----------------------------------
            for k in xrange(self.ip.nz):
                #--------------------------------------
                # In this loop, dz is always a scalar
                #--------------------------------------
                if (SCALAR_dz):    
                    dz = self.ip.dz
                else:    
                    dz = self.ip.dz[k]
                if (dim_q == 3):    
                    qq = self.ip.q[k,:,:]
                else:    
                    qq = self.ip.q[k]
                if (dim_qi == 3):    
                    qqi = self.ip.qi[k,:,:]
                else:    
                    qqi = self.ip.qi[k]
                #----------------------------------------
                # Get increase in q over initial values
                # (but q may have decreased due to ET)
                #----------------------------------------
                dq = (qq - qqi)   #(grid or scalar)
                da
                SCALAR_dq = (size(dq) == 1)
                if (SCALAR_da and SCALAR_dq):    
                    dm = dq * (dz * da * n_pixels)      #(da is a SCALAR)
                else:    
                    dm = sum(double(dq * (da * dz)))    #(da is a GRID)
                vol_stored += dm
            mass_error = float64(100) * (vol_IN - vol_stored) / vol_IN
            err_str = ('%6.2f' % mass_error) + ' % '
            TF_Print('Volume stored      = ' + TF_String(vol_stored) + ' [m^3]')
            TF_Print('Volume error       = ' + err_str)
        TF_Print(' ')

    #   print_mass_balance_report()
    #-------------------------------------------------------------

    


